package refactoringtestplugin;

import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.IPackageFragmentRoot;
import org.eclipse.jdt.core.IProblemRequestor;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.compiler.IProblem;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.PackageDeclaration;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.ltk.core.refactoring.CheckConditionsOperation;
import org.eclipse.ltk.core.refactoring.CreateChangeOperation;
import org.eclipse.ltk.core.refactoring.PerformChangeOperation;
import org.eclipse.ltk.core.refactoring.Refactoring;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;
import org.eclipse.ltk.core.refactoring.RefactoringStatusEntry;

import refactoringtestplugin.util.JavaProjectHelper;
import org.testorrery.IGenerator;

public abstract class RefactoringTest {

	/**
	 * Flag indicating that the runner should check generated compilation units 
	 * for compilation before running refactorings or outputting files.
	 */
	private static final boolean ENABLE_COMPILE_CHECK = true;
	/**
	 * Flag indicating that the runner should run the refactoring specified by
	 * the concrete class. 
	 */
	private static final boolean ENABLE_REFACTORING_EXECUTION = true;
	/**
	 * Flag indicating that the runner should output files to the directory
	 * specified by {@link #getLogDirPath()}
	 */
	private static final boolean ENABLE_FILE_OUTPUT = true;
	/**
	 * The maximum number of compilation units to generate or null for unbounded
	 */
	private static final Integer MAX_GENERATIONS = null;
	
	/**
	 * The root directory in which generated compilation units and the refactoring results 
	 * will be saved.  
	 * 
	 * The returned value is defined by the derived class.  In general it has 
	 * the following form: TEMP_DIR/{refactoring}/{generator}/ where 
	 * {refactoring} is the package name for a particular refacotring and 
	 * {generator} is the specific generator being used to create input data. 
	 * This value will be passed into the {@link TestLogger} that creates the
	 * test-by-test directory structure.
	 */
	protected abstract String getLogDirPath();
	/** 
	 * The two-to-three sentence description of the type of refactoring being performed and the
	 * input used.
	 */
	public abstract String getDescription();
	/**
	 * The generator that provides the inputs for the derived class' test
	 */
	protected abstract IGenerator<CompilationUnit> getCUGen();
	/**
	 * @return the {@link Refactoring} to perform
	 */
	protected abstract Refactoring getRefactoring(ICompilationUnit icu) throws JavaModelException;
	
	public void runTests() throws Exception {
		// Bit of a hack to get the names. 
		// Assumes this class is the generator and superclass is the refactoring.
		String refactoringName = this.getClass().getSuperclass().getSimpleName().replace("Test", "");
		String generatorName = this.getClass().getSimpleName();
		TestLogger logger;
		if (ENABLE_FILE_OUTPUT) {
			logger = new TestLogger(refactoringName, generatorName, getLogDirPath());
		}
		else {
			logger = new TestLogger(refactoringName, generatorName);
		}
		
		ProblemReport problemReport;		
		ICompilationUnit icu;
		
		logger.start(this);
		for (CompilationUnit cu : getCUGen()) {	
			if (generationIsComplete(logger.getGeneratedCount())) break;			
			
			createProject();				
				
			String cuName = getPrimaryTypeName(cu);
			logger.logGenerated(cu, cuName);
			if (ENABLE_COMPILE_CHECK || ENABLE_REFACTORING_EXECUTION) {
				IPackageFragmentRoot packageRoot = createPackageRoot(project);				
				icu = createICompilationUnit(packageRoot, cu, cuName);
			
				if (ENABLE_COMPILE_CHECK) {
					problemReport = checkCompilationProblems(icu);
					if (problemReport.hasProblems()) {
						logger.logPreRefactorNotCompile(icu, problemReport);					
						tearDownProject(); continue;
					}
				}					
				if (ENABLE_REFACTORING_EXECUTION) {
					Refactoring refactoring = getRefactoring(icu); 
					problemReport = checkRefactoringApplies(refactoring);
					if (problemReport.hasProblems()) {
						logger.logRefactoringInapplicable(problemReport);										
						tearDownProject(); continue;
					}				
	
					problemReport = performRefactoring(refactoring);
					if (problemReport.hasProblems()) {
						logger.logRefactoringExecutionErrors(problemReport);
						tearDownProject(); continue;
					}
					
					logger.logPostRefactoring(packageRoot);					

					if (ENABLE_COMPILE_CHECK) {
						problemReport = checkCompilationProblems(packageRoot);
						if (problemReport.hasProblems()) {
							logger.logPostRefactorNotCompile(problemReport);					
							tearDownProject(); continue;
						}
					}
					if (!icu.exists()) {
						// if the refactoring changed or deleted the ICompilationUnit
						icu = getPostRefactoringICompilationUnit();
					}	
					problemReport = checkCustomPostconditions(icu);
					if (problemReport.hasProblems()) {
						logger.logPostconditionFailure(problemReport);
						tearDownProject(); continue;
					}
					logger.logSuccessfulRefactoring();
				}
			}
			tearDownProject();
		}
		logger.stop();
	}

	private boolean generationIsComplete(int generatedCount) {
		return MAX_GENERATIONS != null && generatedCount >= MAX_GENERATIONS;
	}

	private ProblemReport checkRefactoringApplies(Refactoring refactoring) throws CoreException {
		ProblemReport pr = new ProblemReport();		
		RefactoringStatus status = refactoring.checkAllConditions(new NullProgressMonitor());
		if (status.getSeverity() != RefactoringStatus.OK) {
			for (RefactoringStatusEntry entry : status.getEntries()) {
				pr.addProblem(entry.getMessage());
			}
		}
		return pr;
	}
	
	private IJavaProject project;	
	public void createProject() throws Exception {	
		project = JavaProjectHelper.createJavaProject("TestProject"+System.currentTimeMillis(), "bin");
		JavaProjectHelper.addRTJar(project);
	}
	public IJavaProject getProject() {
		return project;
	}
	public void tearDownProject() throws CoreException {
		JavaProjectHelper.delete(project);
	}
	
	protected IPackageFragmentRoot createPackageRoot(IJavaProject project) throws CoreException {
		String packageRootName = "src";
		return JavaProjectHelper.addSourceContainer(project, packageRootName);	
	}
	
	protected ICompilationUnit createICompilationUnit(IPackageFragmentRoot packFragRoot, CompilationUnit cu, String cuName) throws JavaModelException {
		String packageName = getPackageNameFrom(cu);
		IPackageFragment packFrag = packFragRoot.createPackageFragment(packageName, true, null);	
		
		ICompilationUnit icu = packFrag.createCompilationUnit(
				cuName + ".java", // CU name
				cu.toString(), // CU source
				true, // overwrite if CU exists
				new NullProgressMonitor()); // ignore progress
		return icu;
	}

	protected String getPackageNameFrom(CompilationUnit cu) {
		PackageDeclaration packDecl = cu.getPackage();
		if (packDecl == null) {
			return "";
		}
		return packDecl.getName().toString();
	}
	
	private ProblemReport checkCompilationProblems(IPackageFragmentRoot packageRoot) throws JavaModelException {
		ProblemReport problemReport = new ProblemReport();
		IPackageFragment packFrag;
		for (IJavaElement element : packageRoot.getChildren()) {
			if (element instanceof IPackageFragment) {
				packFrag = (IPackageFragment)element;
				for (IJavaElement element2 : packFrag.getChildren()) {
					if (element2 instanceof ICompilationUnit) {
						checkCompilationProblems(problemReport, (ICompilationUnit)element2);
					}
				}
			}
		}
		return problemReport;
	}
	
	protected ProblemReport checkCompilationProblems(ICompilationUnit icu) throws JavaModelException {
		ProblemReport problemReport = new ProblemReport();
		checkCompilationProblems(problemReport, icu);
		return problemReport;
	}
	protected void checkCompilationProblems(final ProblemReport problemReport, ICompilationUnit icu) throws JavaModelException {	
		class CompileStatus implements IProblemRequestor {
			public void acceptProblem(IProblem problem) {
				if (problem.isError()) {
					problemReport.addProblem(problem.getMessage());
				}
			}						
			public void beginReporting() {}
			public void endReporting() {}
			public boolean isActive() {return true;}
		}
		CompileStatus status = new CompileStatus();
		icu.becomeWorkingCopy(status, null);
		icu.discardWorkingCopy();		
	}

	protected String getPrimaryTypeName(CompilationUnit cu) {
		String first = null;
		for (Object type : cu.types()) {
			if (type instanceof TypeDeclaration) {
				String typeName = ((TypeDeclaration)type).getName().toString();
				if (first == null) {
					first = typeName;
				}
				for (Object modifier : ((TypeDeclaration)type).modifiers()) {
					if (((Modifier)modifier).getKeyword() == Modifier.ModifierKeyword.PUBLIC_KEYWORD) {						
						return typeName;
					}
				}
			}
		}
		return first;
	}

	// simplified from org.eclipse.jdt.ui.tests.refactoring.RefactoringTest.performRefactoring(Refactoring ref, boolean providesUndo)
	protected ProblemReport performRefactoring(Refactoring refactoring) throws CoreException {	
		final CreateChangeOperation create= new CreateChangeOperation(
				new CheckConditionsOperation(refactoring, CheckConditionsOperation.ALL_CONDITIONS),
				RefactoringStatus.FATAL);
		PerformChangeOperation perform = new PerformChangeOperation(create);
		
		IWorkspace workspace = ResourcesPlugin.getWorkspace();
		workspace.run(perform, new NullProgressMonitor());
		
		RefactoringStatus conditionCheckingStatus = create.getConditionCheckingStatus();
		ProblemReport problemReport = new ProblemReport();
		if (conditionCheckingStatus.getSeverity() != RefactoringStatus.OK) {
			problemReport.addProblems(conditionCheckingStatus.getEntries());
		}
		return problemReport;
	}
	
	/**
	 * Should be overridden by refactorings that change the target {@link ICompilationUnit}
	 * @return a new {@link ICompilationUnit} if the old one was changed by a refactoring
	 */
	protected ICompilationUnit getPostRefactoringICompilationUnit() {
		throw new RuntimeException("ICompilationUnit should not have changed");
	}
	
	/**
	 * Checks any additional constraints on the refactored compilation unit.
	 * Should be overridden by derived classes that need to specify custom checks.
	 * @throws JavaModelException 
	 * @throws Exception 
	 */
	protected ProblemReport checkCustomPostconditions(ICompilationUnit icu) throws Exception {
		return new ProblemReport();
	}
	
	/**
	 * Utility function that returns the system's temporary directory.
	 */
	public static String getSystemTempDir() {
		String tempdir = System.getProperty("java.io.tmpdir");
		if (tempdir == null) {
			throw new IllegalArgumentException("Temp dir is not specified");
		}
		String separator = System.getProperty("file.separator");
		if (!tempdir.endsWith(separator)) {
			return tempdir + separator;
		}
		return tempdir;
	}
}
