package refactoringtestplugin;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.LinkedList;

import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.IPackageFragmentRoot;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.CompilationUnit;

/**
 * Logs refactoring test results to the console and a directory structure.
 * 
 * The directory structure has the following format:
 * 
 * <pre>
 * {logDir}N/{SUMMARY_FILE_NAME}
 *           {TEST_NAME_BASE}M/
 *                            in/
 *                              {compilationUnit}.java
 *                            out/
 *                               {compilationUnit}.java
 *                               {SUCCESSFUL_REFACTORING_LOG_FILE}
 *                               {PRE_REFACTOR_NOT_COMPILE_LOG_FILE}
 *                               {REFACTORING_INAPPLICABLE_LOG_FILE}
 *                               {REFACTORING_EXECUTION_ERROR_LOG_FILE}
 *                               {POST_REFACTOR_NOT_COMPILE_LOG_FILE}
 *                               {POSTCONDITION_FAILURE_LOG_FILE}
 * </pre>
 * 
 * Where {logDir} is the directory passed into the constructor (usually from 
 * {@link RefactoringTest#getLogDirPath()}), N and M are 
 * unique numbers to ensure that multiple runs do not overwrite each other, 
 * and other curly-braced values are those defined in the class.  
 */
public class TestLogger {	
	private static final String TEST_NAME_BASE = "test";
	private static final String PRE_REFACTOR_NOT_COMPILE_LOG_FILE = "PRE_REFACTOR_NOT_COMPILE";
	private static final String REFACTORING_INAPPLICABLE_LOG_FILE = "REFACTORING_INAPPLICABLE";
	private static final String REFACTORING_EXECUTION_ERROR_LOG_FILE = "REFACTORING_EXECUTION_ERROR";
	private static final String POST_REFACTOR_NOT_COMPILE_LOG_FILE = "POST_REFACTOR_NOT_COMPILE";
	private static final String POSTCONDITION_FAILURE_LOG_FILE = "POSTCONDITION_FAILURE";
	private static final String SUCCESSFUL_REFACTORING_LOG_FILE = "SUCCESSFUL_REFACTORING";
	private static final String SUMMARY_FILE_NAME = "SUMMARY";
	
	private int generatedCount = 0;
	private int preRefactorNotCompileCount = 0;
	private int refactoringInapplicableCount = 0;
	private int refactoringExecutionErrorCount = 0;
	private int postRefactoringNotCompileCount = 0;
	private int postconditionFailureCount = 0;
	private int successfulRefactorings = 0;
	private long start;	
	private long stop;
	
	private File logDir;
	private String refactoringName;
	private String generatorName;
	
	public TestLogger(String refactoringName, String generatorName) {
		this.refactoringName = refactoringName;
		this.generatorName = generatorName;
	}
	
	public TestLogger(String refactoringName, String generatorName, String logDirPath) {
		this(refactoringName, generatorName);
		this.logDir = createUniqueLogDir(logDirPath, 0);
	}

	private File createUniqueLogDir(String logDirPathBase, int suffix) {
		String logDirPath = logDirPathBase + suffix;
		File uniqueLogDir = new File(logDirPath);
		if (uniqueLogDir.exists()) {
			return createUniqueLogDir(logDirPathBase, suffix + 1);
		}
		info("Creating log directory " + uniqueLogDir.getAbsolutePath());
		if (!uniqueLogDir.mkdirs()) {
			throw new RuntimeException(String.format("Log directory %s could not be created", uniqueLogDir.getAbsolutePath()));
		}
		return uniqueLogDir;
	}
	
	public boolean fileOutputIsEnabled() {
		return logDir != null;
	}
	
	public int getGeneratedCount() {
		return generatedCount;
	}
	
	public void start(RefactoringTest test) {		
		info("Starting " + test.getClass().getName().toString());
		info("Description: " + test.getDescription());
		start = System.currentTimeMillis();
	}
	
	public void logGenerated(CompilationUnit cu, String cuName) {
		generatedCount++;
		info("==================================");
		String cuSource = cu.toString();
		info(String.format("Input %d before refactoring:\n%s", generatedCount, cuSource));
		outputCompilationUnitFile("in", cuSource, cuName);
	}

	public void logPreRefactorNotCompile(ICompilationUnit icu, ProblemReport problemReport) {
		preRefactorNotCompileCount++;
		warn("Generated input does not compile");
		
		for (String problem : problemReport.getProblems()) {
			warn(problem);
		}
		outputToLogFile(PRE_REFACTOR_NOT_COMPILE_LOG_FILE, problemReport.getProblems());		
	}

	public void logRefactoringInapplicable(ProblemReport problemReport) {
		refactoringInapplicableCount++;
		warn("Refactoring does not apply");
		
		for (String problem : problemReport.getProblems()) {
			warn(problem);
		}
		outputToLogFile(REFACTORING_INAPPLICABLE_LOG_FILE, problemReport.getProblems());		
	}
	
	public void logRefactoringExecutionErrors(ProblemReport problemReport) {
		refactoringExecutionErrorCount++;
		fail("Error(s) encountered during refactoring execution");
		
		for (String problem : problemReport.getProblems()) {
			fail(problem);
		}
		outputToLogFile(REFACTORING_EXECUTION_ERROR_LOG_FILE, problemReport.getProblems());
	}
	
	public void logPostRefactoring(IPackageFragmentRoot packageRoot) throws JavaModelException {
		info(String.format("Input %d after refactoring", generatedCount));
		IPackageFragment packFrag;
		ICompilationUnit icu;
		for (IJavaElement element : packageRoot.getChildren()) {
			if (element instanceof IPackageFragment) {
				packFrag = (IPackageFragment)element;
				for (IJavaElement element2 : packFrag.getChildren()) {
					if (element2 instanceof ICompilationUnit) {
						icu = (ICompilationUnit)element2;
						String cuSource = icu.getSource();
						String cuName = icu.getElementName();
						info(cuName + "\n" + cuSource);
						outputCompilationUnitFile("out", cuSource, cuName);
					}
				}
			}
			
		}
	}

	public void logPostRefactorNotCompile(ProblemReport problemReport) {
		postRefactoringNotCompileCount++;
		fail("Unit does not compile after refactoring");
	
		for (String problem : problemReport.getProblems()) {
			fail(problem);
		}
		outputToLogFile(POST_REFACTOR_NOT_COMPILE_LOG_FILE, problemReport.getProblems());	
	}

	public void logPostconditionFailure(ProblemReport problemReport) {
		postconditionFailureCount++;
		fail("Custom post-condition failure(s) found");
		
		for (String problem : problemReport.getProblems()) {
			fail(problem);
		}
		outputToLogFile(POSTCONDITION_FAILURE_LOG_FILE, problemReport.getProblems());	
	}
	
	public void logSuccessfulRefactoring() {
		outputToLogFile(SUCCESSFUL_REFACTORING_LOG_FILE, null);
		successfulRefactorings++;
	}
	
	public void stop() {
		stop = System.currentTimeMillis();		
		outputSummary();
	}

	private static class Result {
		public String name;
		public String value;
		public Result(String name, Object value) {
			this.name = name;
			this.value = value.toString();
		}
	}
	private void outputSummary() {
		LinkedList<Result> results = new LinkedList<Result>();		
		results.add(new Result("Refactoring", refactoringName));
		results.add(new Result("Primary Generator", generatorName));
		results.add(new Result("Total Generated", generatedCount));
		results.add(new Result("Time", (stop - start) / 1000F));
		results.add(new Result("Compilable Inputs", generatedCount - preRefactorNotCompileCount));
		//results.add(new Result("Pre-Refactoring Compilation Failures", preRefactorNotCompileCount));
		results.add(new Result("WarningStatus (Ecl)", refactoringInapplicableCount));
		results.add(new Result("WarningStatus (NB)", "\\Fix{NBWS}"));
		//results.add(new Result("Successful Refactorings", successfulRefactorings));
		//results.add(new Result("Refactoring Execution Errors", refactoringExecutionErrorCount));
		results.add(new Result("DoesNotCompile (Ecl)", postRefactoringNotCompileCount));
		results.add(new Result("DoesNotCompile (NB)", "\\Fix{NBDNC}"));
		results.add(new Result("Custom/Inverse", postconditionFailureCount));
		results.add(new Result("Differential", "\\Fix{Diff}"));
		results.add(new Result("Bugs Reported (Ecl)", "\\Fix{Ebugs}"));
		results.add(new Result("Bugs Reported (NB)", "\\Fix{NBbugs}"));
		
		outputConsoleSummary(results);
		outputSummaryLogFile(results);
	}

	private void outputConsoleSummary(LinkedList<Result> results) {
		for (Result result : results) {
			System.out.print(result.name);
			System.out.print(": ");
			System.out.println(result.value);
		}
	}

	public void info(String info) {
		System.out.println("INFO: " + info);
	}
	public void warn(String warning) {
		System.out.println("WARNING: " + warning);
	}
	public void fail(String failure) {
		System.out.println("FAILURE: " + failure);
	}
	
	protected void outputToLogFile(String logFileName, Iterable<String> messages) {
		if (!fileOutputIsEnabled()) {
			return;
		}
		File logFile = new File(getInOutDir("out"), logFileName);
		try {				
			if (!logFile.createNewFile()) {
				throw new RuntimeException(String.format("Log file %s could not be created", logFile.getAbsolutePath()));
			}
			FileWriter fw = new FileWriter(logFile);
			if (messages != null) {
				for (String message : messages) {					
					fw.write(message);
					fw.write('\n');
				}
			}
			fw.flush();
			fw.close();

		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	protected void outputCompilationUnitFile(String inOutDirName, String source, String cuName) {
		if (!fileOutputIsEnabled()) {
			return;
		}
		
		File inOutDir = getInOutDir(inOutDirName);		
		if (!cuName.endsWith(".java")) {
			cuName = cuName + ".java";
		}
		File cuFile = new File(inOutDir, cuName); // ../<compilation_unit_name>.java
		try {
			if (!cuFile.createNewFile()) {
				throw new RuntimeException(String.format("Compilation unit file %s could not be created", cuFile.getAbsolutePath()));			
			}
			FileWriter fw = new FileWriter(cuFile);
			fw.write(source);
			fw.flush();
			fw.close();
		} 
		catch (Exception e) {
			throw new RuntimeException(e); //HACK
		} 			
	}

	protected File getInOutDir(String inOutDirName) {
		if (!fileOutputIsEnabled()) {
			return null;
		}
		File inOutDir = new File(getTestDirectory(), inOutDirName); //  <temp_dir>/<refactoring>/<test_dir>/(in|out)
		if (!inOutDir.exists()) {
			if (!inOutDir.mkdirs()) {
				throw new RuntimeException(String.format("In/Out directory %s could not be created", inOutDir.getAbsolutePath()));
			}
		}
		return inOutDir;
	}
	
	protected File getTestDirectory() {
		if (!fileOutputIsEnabled()) {
			return null;
		}
		File curTestDir = new File(logDir, TEST_NAME_BASE + generatedCount);
		if (!curTestDir.exists()) {
			if (!curTestDir.mkdirs()) {
				throw new RuntimeException(String.format("Test directory %s could not be created", curTestDir.getAbsolutePath()));
			}
			info("Test directory created at " + curTestDir.getAbsolutePath());
		}			
		return curTestDir;
	}
	

	private void outputSummaryLogFile(Iterable<Result> results) {
		if (!fileOutputIsEnabled()) {
			return;
		}
		File summaryFile = new File(logDir, SUMMARY_FILE_NAME);
		try {
			if (!summaryFile.createNewFile()) {
				throw new RuntimeException("Unable to create summary file");
			}
			FileWriter fw = new FileWriter(summaryFile);
			writeHeaders(results, fw);
			writeValues(results, fw);
			fw.flush();
			fw.close();
		} 
		catch (IOException e) {
			throw new RuntimeException(e); //HACK
		}
	}

	private void writeHeaders(Iterable<Result> results, FileWriter fw) throws IOException {
		fw.write("%%");
		boolean first = true;			
		for (Result result : results) {
			if (!first) {
				fw.write(" & ");
			}
			fw.write("\\RHD{"); // the rotated header command in the TeX file
			fw.write(result.name);
			fw.write("}");
			first = false;
		}
		fw.write('\n');
	}

	private void writeValues(Iterable<Result> results, FileWriter fw) throws IOException {
		Boolean first2 = true;
		for (Result result : results) {
			if (!first2) {
				fw.write(" & ");
			}
			fw.write(result.value);
			first2 = false;
		}
		fw.write("\\\\\n");
	}
}