package org.testorrery;
import java.util.Iterator;
import java.util.LinkedList;

/**
 * Iterates through all values generated by a set of child generators
 *
 * @param <ValueType> The type of value this generator returns
 */
public class Chain<ValueType> extends Generator<ValueType> {
	
	/**
	 * the list of generators to be iterated through in sequence 
	 */
	private LinkedList<IGenerator<ValueType>> generators; 
	/**
	 * the iterator of generators being iterated against 
	 */
	private Iterator<IGenerator<ValueType>> iterator;
	/**
	 * the current generator being iterated against
	 */
	private IGenerator<ValueType> currentGenerator;	
	
	public Chain() {	
		super();
		generators = new LinkedList<IGenerator<ValueType>>();		
	}
	
	public Chain(ValueType...values) {
		this();
		for (ValueType value : values) {
			this.add(value);
		}
	}
	
	public Chain(IGenerator<ValueType>... generators) {
		this();
		for (IGenerator<ValueType> gen : generators) {
			this.add(gen);
		}	
	}
	
	/**
	 * Add a generator to the end of the sequence
	 */
	public <GeneratorType extends IGenerator<ValueType>> void add(GeneratorType generator) {
		generators.add(generator);
		reset(); 
	}
	
	/**
	 * Add a value to the end of the sequence
	 * @param value The value to add
	 */
	public void add(ValueType value) {
		add(new Literal<ValueType>(value));
	}
	
	/**
	 * 
	 * @return The generator in the sequence currently being iterated against
	 */
	public IGenerator<ValueType> currentGenerator() {
		return currentGenerator;
	}
	
	@Override
	public boolean hasNext() {
		if (currentGenerator != null && currentGenerator.hasNext()) {
			return true;
		}
		// loop over empty generators
		while (iterator != null && iterator.hasNext()) {
			currentGenerator = iterator.next();
			if (currentGenerator.hasNext()) {
				return true;
			}
		}
		return false;
	}

	@Override
	protected ValueType generateNext() {
		// the current generator should be set in canGenerate() via hasNext(). If it was not, then throw an error
		if (!hasNext() || currentGenerator == null || !currentGenerator.hasNext()) {
			throw new IterationException();
		}
		else {
			return currentGenerator.next();
		}
	}
	
	public void reset() {
		super.reset();
		iterator = null;
		currentGenerator = null; 
		
		if (!generators.isEmpty()) {
			iterator = generators.iterator();				
			for (IGenerator gen : generators) {
				gen.reset();
			}
		}
	}

}
