package org.testorrery;

/**
 * A Literal generator that chooses one of several possible values as its return value.
 * Used for random testing.
 * 
 * Note that setChangeOnReset can be set to true to allow the generator to choose
 *  a new value from the given choices when <code>reset</code> is called. 
 * @param <ValueType> The type of value to return
 */
public class Choose<ValueType> extends Literal<ValueType> {	
	private java.util.Random seed;
	private IGenerator<ValueType> valueGenerator;
	private int maxIterations;
	
	/**
	 * Indicates that this generator should choose a new value when 
	 *  <code>reset</code> is called
	 */
	private boolean changeValueOnReset;
	
	/**
	 * Randomly chooses one of the given values 
	 */
	public Choose(ValueType... possibleValues) {
		this(new java.util.Random(), possibleValues);
	}
	
	/**
	 * Randomly chooses one of the given values 
	 */
	public Choose(java.util.Random seed, ValueType... possibleValues) {	
		this(seed, possibleValues.length, new Chain<ValueType>(possibleValues));
	}	
	
	/**
	 * Chooses a random value from the generator.
	 * Note that this constructor must run through the generator twice. If you 
	 *  know the maxIterations, it will be less expensive to call the appropriate
	 *  overloaded constructor.
	 */
	public Choose(IGenerator<ValueType> valueGenerator) {
		this(new java.util.Random(), valueGenerator);
	}
	
	/**
	 * Chooses a random value from the generator
	 * Note that this constructor must run through the generator twice. If you 
	 *  know the maxIterations, it will be less expensive to call the appropriate
	 *  overloaded constructor.
	 */
	public Choose(java.util.Random seed, IGenerator<ValueType> valueGenerator) {
		this(seed, getMaxIterations(valueGenerator), valueGenerator);
	}
	
	/**
	 * Iterates the given generator a random number of times between 
	 *  1 and maxIterations (inclusive), then uses the generator's current 
	 *  value as this generator's value.
	 */
	public Choose(int maxIterations, IGenerator<ValueType> valueGenerator) {
		this(new java.util.Random(), maxIterations, valueGenerator);
	}
	
	/**
	 * Iterates the given generator a random number of times between 
	 *  1 and maxIterations (inclusive), then uses the generator's current 
	 *  value as this generator's value.
	 */
	public Choose(java.util.Random seed, int maxIterations, IGenerator<ValueType> valueGenerator) {
		super(getFromGenerator(seed, valueGenerator, maxIterations));
		this.seed = seed;
		this.valueGenerator = valueGenerator;	
		this.maxIterations = maxIterations;
		
		this.changeValueOnReset = false;
	}
	
	/**
	 * Utility function that returns the number of times the generator can iterate.
	 * Public only for testing. Do not use.
	 */
	public static int getMaxIterations(IGenerator valueGenerator) {
		// figure out the maximum number of iterations by running the generator to exhaustion.
		//  This may sound expensive, but getFromGenerator needs to generate from the beginning anyway
		int maxIterations = 0;
		valueGenerator.reset();
		while(valueGenerator.hasNext()) {
			valueGenerator.next();
			maxIterations++;
		}
		return maxIterations;
	}
	
	/**
	 * Utility function that returns a random value from the generator.
	 */
	private static <ValueType> ValueType getFromGenerator(java.util.Random seed, IGenerator<ValueType> valueGenerator, int maxIterations) {
		if (maxIterations <= 0) {
			throw new IllegalArgumentException("maxIterations must be > 0");
		}
		int useIndex = seed.nextInt(maxIterations);
		valueGenerator.reset();
		valueGenerator.next(); // make sure the generator is not reset
		while (useIndex-- > 0) {
			if (!valueGenerator.hasNext()) {
				throw new IllegalArgumentException("Invalid maxIterations");
			}
			valueGenerator.next();
		}
		return valueGenerator.current();
	}
	
	/**
	 * Indicates whether or not this generator should choose a new
	 *  random value when <code>reset</code> is called
	 */
	public void setChangeOnReset(boolean shouldChange) {
		changeValueOnReset = shouldChange;
	}	
	
	@Override
	public void reset() {
		super.reset();
		if (changeValueOnReset) {
			// set this value to the another random value from the generator
			super.setInnerValue(getFromGenerator(seed, valueGenerator, maxIterations));
		}
	}
}
