package org.testorrery;

import org.testorrery.link.IIterationLink;

/**
 * Abstract base class for generators that have one or more child generators.
 * The derived class need only supply an {@link IIterationLink}, and this
 *  base class will handle the iteration. 
 * 
 * @param <ValueType> The type of value generated
 */
public abstract class CompositeGenerator<ValueType> extends Generator<ValueType> {

	private IIterationLink iterator;	
	
	public CompositeGenerator() {
		super();		
	}

	/**
	 * Returns the derived class' iterator
	 */
	public IIterationLink getIterator() {
		if (iterator == null) {	
			iterator = getInitializedIterator();
		}
		return iterator;
	}
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean hasNext() {
		return getIterator().hasNext();
	}

	/**
	 * Iterates the iterator then delegates to the derived class to generate
	 *  the object from the current value(s) of the iterated generator(s).
	 */
	@Override
	protected ValueType generateNext() {
		if (!getIterator().hasNext()) {
			throw new IterationException();
		}
		getIterator().next();
		return generateCurrent();
	}
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public void reset() {
		super.reset();
		getIterator().reset();
	}	
	
	/**
	 * Generate a value from the current values of the child generators
	 */
	protected abstract ValueType generateCurrent();
	
	/**
	 * Return an iterator initialized with all child generators
	 */
	protected abstract IIterationLink getInitializedIterator();	
}
