package org.testorrery;

import java.util.LinkedList;

/**
 * A generator that will only return those values from a sub-generator that meet 
 *  criteria provided by a list of {@link IPredicate}s.
 */
public class Filter<ValueType> extends Generator<ValueType> {	
	private LinkedList<IPredicate<ValueType>> predicates;	
	private IGenerator<ValueType> innerGenerator;
	private ValueType generated;
	private boolean hasGenerated;
	
	public Filter(IGenerator<ValueType> innerGenerator) {
		predicates = new LinkedList<IPredicate<ValueType>>();
		this.innerGenerator = innerGenerator;
		generated = null;
		hasGenerated = false;
	}
	
	/**
	 * Registers a predicate that defines a test that generated objects must satisfy
	 * @param predicate A predicate to check
	 */
	public void registerPredicate(IPredicate<ValueType> predicate) {
		predicates.add(predicate);
	}
	
	/**
	 * Checks if <code>value</code> satisfies all registered predicates
	 * @param value The value to check 
	 * @return true if <code>value</code> satisfies all registered predicates
	 */
	private boolean satisfiesPredicates(ValueType value) {
		for (IPredicate<ValueType> predicate : predicates) {
			if (!predicate.evaluate(value)) {
				return false;
			}
		}
		return true;
	}
	
	@Override
	public boolean hasNext() {
		// generate values until one satisfies the criteria
		while (innerGenerator.hasNext() && !hasGenerated) {
			generated = innerGenerator.next();
			if (satisfiesPredicates(generated)) {
				hasGenerated = true;
			}
		}
		return hasGenerated;
	}
	
	@Override
	protected ValueType generateNext() {
		if (!hasGenerated && !hasNext()) {
			throw new IterationException();
		}
		hasGenerated = false;
		return generated;
	}
	
	@Override
	public void reset() {
		super.reset();
		innerGenerator.reset();
		generated = null;
		hasGenerated = false;
	}
}
