package org.testorrery;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;

import org.testorrery.link.CombinatoricLink;
import org.testorrery.link.GeneratorLink;
import org.testorrery.link.IIterationLink;
import org.testorrery.link.SequentialLink;

/**
 * A generator that when given several prototype objects will 
 * produce objects containing the prototypes' values.
 */
public class PrototypeGenerator<T> extends CompositeGenerator<T> {

	private Class<? extends T> clazz;
	private HashMap<String, PrototypeFieldAccessor<T>> fields = new HashMap<String, PrototypeFieldAccessor<T>>();

	public PrototypeGenerator(Class<? extends T> clazz) {
		this.clazz = clazz;	
		findAccessors(clazz);
	}
	
	public Class<? extends T> getPrototypeClass() {
		return clazz;
	}

	private void findAccessors(Class<? extends T> clazz) {
		findPublicFields(clazz);
		findAccessorMethods(clazz);		
		if (fields.size() == 0) {
			throw new RuntimeException("No public fields or field accessors found");
		}
	}

	private void findPublicFields(Class<? extends T> clazz) {
		for (Field field : clazz.getFields()) {
			fields.put(field.getName(), createPublicFieldAccessor(field));
		}
	}

	protected PrototypeFieldAccessor<T> createPublicFieldAccessor(Field field) {
		return new PublicField(field);
	}
	
	private void findAccessorMethods(Class<? extends T> clazz) {
		HashMap<String, Method> getters = new HashMap<String, Method>();
		HashMap<String, Method> setters = new HashMap<String, Method>();
		for (Method method : clazz.getMethods()) {
			if (method.getName().startsWith("get")) {
				getters.put(fieldNameFromAccessor(method), method);
			}
			else if (method.getName().startsWith("set")) {
				setters.put(fieldNameFromAccessor(method), method);
			}
		}
		for (String fieldName : getters.keySet()) {
			if (setters.containsKey(fieldName)) {
				fields.put(fieldName, createMethodAccessor(getters.get(fieldName), setters.get(fieldName)));
			}
		}
	}

	protected PrototypeFieldAccessor<T> createMethodAccessor(Method getter, Method setter) {
		return new AccessorMethod(getter, setter);
	}

	private String fieldNameFromAccessor(Method method) {
		String fieldBase = method.getName().substring(3);
		return fieldBase.substring(0, 1).toLowerCase() + fieldBase.substring(1);
	}
	
	public void addPrototype(T proto) {
		try {		
			for (PrototypeFieldAccessor<T> field : fields.values()) {
				field.addPrototype(proto);
			}
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}
	
	public void setCombinatoric(String fieldName) throws SecurityException, NoSuchFieldException {
		fields.get(fieldName).setIsCombinatoric();		
	}
	
	@Override
	protected IIterationLink getInitializedIterator() {
		IIterationLink seqLink = null;
		IIterationLink combLink = null;
		GeneratorLink leaf;
		// create a sequential link containing all sequential
		// fields and a combinatoric link containing all combinatoric
		// fields and combine them later.
		for (PrototypeFieldAccessor field : fields.values()) {
			leaf = field.getGeneratorLink();
			if (field.isCombinatoric()) {
				if (combLink == null) {
					combLink = leaf;
				}
				else {
					combLink = new CombinatoricLink(leaf, combLink);
				}
			}
			else if (seqLink == null) {
				seqLink = leaf;
			}
			else {
				seqLink = new SequentialLink(leaf, seqLink);
			}
		}		
		if (combLink == null) {
			return seqLink;
		}
		if (seqLink == null) {
			return combLink;
		}
		return new CombinatoricLink(combLink, seqLink);
	}

	@Override
	protected T generateCurrent() {
		try {
			T toReturn = newInstance();
			for (PrototypeFieldAccessor<T> field : fields.values()) {
				field.fillCurrentValue(toReturn);				
			}		
			return toReturn;
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	protected T newInstance() throws InstantiationException, IllegalAccessException {
		return clazz.newInstance();
	}
	
	private class PublicField extends PrototypeFieldAccessor<T> {
		private Field field;

		public PublicField(Field field) {
			this.field = field;
		}

		@Override
		protected Object getValue(T instance) throws IllegalArgumentException, IllegalAccessException {
			return field.get(instance);
		}

		@Override
		public void setValue(T instance, Object value) throws IllegalArgumentException, IllegalAccessException {
			field.set(instance, value);
		}
	}

	private class AccessorMethod extends PrototypeFieldAccessor<T> {
		private Method getter;
		private Method setter;

		public AccessorMethod(Method getter, Method setter) {
			this.getter = getter;
			this.setter = setter;
		}

		@Override
		protected Object getValue(T instance) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException {
			return getter.invoke(instance);
		}

		@Override
		public void setValue(T instance, Object value) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException {
			setter.invoke(instance, value);
		}
	}
}


