package org.testorrery;

/**
 * A generator that wraps a given generator and repeats its iteration up to a given number of times
 *
 * @param <ValueType> The type of value the wrapped generator (and this repeater) generates
 */
public class Repeater<ValueType> extends Generator<ValueType> {

	private IGenerator<ValueType> generator;
	private int repetitions;
	private int curRepetitions;
	
	public Repeater(IGenerator<ValueType> generator, int repetitions) {
		super();
		if (repetitions <= 0) {
			throw new IllegalArgumentException("repetitions must be > 0");
		}
		this.generator = generator;
		this.repetitions = repetitions;
		this.curRepetitions = 0;
	}
	
	@Override
	public boolean hasNext() {
		// if it has more repetitions or its on the last repetition and can generate more
		return curRepetitions < repetitions - 1 
		|| (curRepetitions == repetitions - 1 && generator.hasNext());
	}

	@Override
	protected ValueType generateNext() {
		if (curRepetitions >= repetitions) {
			throw new IterationException();
		}		
		// if the generator has completed, start another repetition
		if (!generator.hasNext()) {
			curRepetitions++;
			generator.reset();
		}

		return generator.next();
	}
	
	@Override
	public void reset() {
		super.reset();
		generator.reset();
		curRepetitions = 0;
	}

}
