package org.testorrery;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Hashtable;

import org.testorrery.link.CombinatoricLink;
import org.testorrery.link.IIterationLink;
import org.testorrery.link.LinkCreator;

/**
 * Generator that allows one to define a custom generator at runtime.
 * 
 * <p>Say one has the following simple class:</p>
 * 
 * <pre>
 * class SampleClass {
 *   public int field1; 
 *   private String field2;
 *   private String field3;
 *   
 *   public setField2(int field2) {
 *     this.field2 = field2;
 *   }
 * }
 * </pre>
 * 
 * <p>One can create a RuntimeGenerator for this class in the following way:</p>
 * 
 * <pre>
 * RuntimeGenerator&lt;SampleClass&gt; rtGen = new RuntimeGenerator&lt;SampleClass&gt;(SampleClass.class);
 * rtGen.setGenerator("field1", new Chain(1, 2, 3)); // sets public member variable explicitly 
 * rtGen.setGenerator("field2", new Chain("ABC", "DEF");  // uses accessor method
 * //rtGen.setGenerator("field3", new Literal("GHI")); // throws exception because field3 is private and has no accessor
 * </pre>
 * 
 * <p>A RuntimeGenerator can be iterated just like any other generator:</p>
 * 
 * <pre>
 * for (SampleClass sample : rtGen) {
 *   ...
 * }
 * </pre>
 * 
 * @param <ValueType> The type of value returned from the generator
 */
public class RuntimeGenerator<ValueType> extends CompositeGenerator<ValueType> {
	private Class<? extends ValueType> classDef;
	private Hashtable<Field, IGenerator> fieldGenerators;
	private Hashtable<Method, IGenerator> setterGenerators;
	
	public RuntimeGenerator(Class<? extends ValueType> classDef) {
		fieldGenerators = new Hashtable<Field, IGenerator>();
		setterGenerators = new Hashtable<Method, IGenerator>();
		this.classDef = classDef;
	}
	
	/**
	 * Links the given field to the value returned from the given generator on each iteration.
	 * The generated value will be passed to a setter method if one exists. Otherwise, the
	 *  field will be set directly. 
	 * 
	 * @param fieldName The name of the field whose value to set
	 * @param generator The generator whose value will be set to the field on each iteration.
	 */
	public void setGenerator(String fieldName, IGenerator generator) throws SecurityException, NoSuchFieldException {	

		// look for setter method
		//  cannot use getMethod() because we do not know what type the single argument should be
		String setterName = "set" + Character.toUpperCase(fieldName.charAt(0)) + (fieldName.substring(1, fieldName.length()));
		for (Method method : classDef.getMethods()) {
			if (method.getName().equals(setterName) && method.getParameterTypes().length == 1) {
				setterGenerators.put(method, generator);
				return;
			}
		}
				
		// no setter method, so use field itself
		Field field = classDef.getField(fieldName); // throws stuff		
		fieldGenerators.put(field, generator);	
	}

	@Override
	protected IIterationLink getInitializedIterator() {
		LinkCreator lc = new LinkCreator(new CombinatoricLink());
		// field generators
		for (IGenerator gen : fieldGenerators.values()) {
			lc.add(gen);
		}
		// generators for setter methods
		for (IGenerator gen : setterGenerators.values()) {
			lc.add(gen);
		}
		return lc.getLink();
	}
	
	@Override
	protected ValueType generateCurrent() {
		ValueType toReturn;
		
		try {
			toReturn = classDef.newInstance();
			
			// set member variables to generator values
			for (Field field : fieldGenerators.keySet()) {			
				field.set(toReturn, fieldGenerators.get(field).current());				
			}		
			// pass generator values to setter methods
			for (Method setter : setterGenerators.keySet()) {
				setter.invoke(toReturn, setterGenerators.get(setter).current());
			}
		} 
		catch (InstantiationException e) {
			throw new GenerationException(e);
		} 
		catch (IllegalAccessException e) {
			throw new GenerationException(e);
		}
		catch (SecurityException e) {
			throw new GenerationException(e);
		} 
		catch (IllegalArgumentException e) {
			throw new GenerationException(e);
		} 
		catch (InvocationTargetException e) {
			throw new GenerationException(e.getCause());
		} 
		
		return toReturn;
		
	}

}
