package org.testorrery.link;

import java.util.StringTokenizer;

import org.testorrery.IGenerator;

class LinkExpressionParser {
	
	public final static char SEQUENTIAL_OPERATOR = '+';
	public final static char COMBINATORIC_OPERATOR = '*';
		
	public static IIterationLink parse(
			String expression, 
			IGenerator... generators) throws LinkExpressionParseException {
		return parse(
				new StringTokenizer(expression, "$() \n\r\f\t" + SEQUENTIAL_OPERATOR + COMBINATORIC_OPERATOR, true), 
				false,
				generators.clone()); // copy to keep from clobbering the array
	}
	
	/**
	 * Recursive descent parser that returns a root Link.
	 * Creates a Link that will iterate the given generators in the manner specified by the expression.
	 * The expression acts like a format string with the following terms:
	 * <ul>
	 *  <li>$n corresponds to the nth generator in the list (zero-based). Note that a particular generator can only be used once in an expression.</li>
	 *  <li>+ links two generators sequentially (that is, the expression will iterate as long as one of the generators has not yet completed).</li>
	 *  <li>* links two generators combinatorically.</li>
	 *  <li>( and ) affect how expressions are linked</li>
	 * </ul>
	 * Order of operations follows normal arithmetic operators. 
	 * @param expression The expression specifying how generators are linked
	 * @param generators The generators to link
	 * @return A link that will iterate the given generators as specified by the expression
	 */	
	private static IIterationLink parse(
			StringTokenizer tokens, 
			boolean expectParen, 
			IGenerator[] generators) throws LinkExpressionParseException {
		String curToken = getNonblankToken(tokens);
		IIterationLink left;
		IIterationLink right;
		String op;
		
		if (curToken.equals("$")) {
			// generator reference
			curToken = getNonblankToken(tokens); // expect the index of the generator
			try {
				int genIndex = Integer.parseInt(curToken);
				if (generators[genIndex] != null) {
					left = new GeneratorLink(generators[genIndex]);
					generators[genIndex] = null; // to prevent duplicate indices in expression
				}
				else {
					throw new LinkExpressionParseException("Duplicate generator index: " + genIndex);
				}		
			}
			catch (ArrayIndexOutOfBoundsException e) {
				throw new LinkExpressionParseException("Invalid generator index");
			}
			
		}
		else if (curToken.equals("(")) {
			// parenthesized expression
			left = parse(tokens, true, generators);
		}
		else {
			throw new LinkExpressionParseException("Invalid token \"" + curToken + "\"");
		}		
		
		if (!tokens.hasMoreTokens()) {
			if (expectParen) {
				throw new LinkExpressionParseException("Unbalanced parentheses");
			}
			else {
				return left;
			}
		}
		
		op = getNonblankToken(tokens);
		
		if (op.equals(")")) {
			if (expectParen) {
				return left;
			}
			else {
				throw new LinkExpressionParseException("Unbalanced parentheses");
			}
		}
		
		right = parse(tokens, expectParen, generators);
		
		if (op.equals(String.valueOf(SEQUENTIAL_OPERATOR))) {
			return new SequentialLink(left, right);
		}
		else if (op.equals(String.valueOf(COMBINATORIC_OPERATOR))) {
			return new CombinatoricLink(left, right);
		}
		else {
			throw new LinkExpressionParseException("Invalid operator \"" + op + "\"");
		}	
	}
	
	private static String getNonblankToken(StringTokenizer tokens) throws LinkExpressionParseException {
		String token = "";
		// loop until first nonblank token
		while (token.equals("")) {
			if (tokens.hasMoreTokens()) {
				token  = tokens.nextToken().trim();
			}
			else {
				throw new LinkExpressionParseException("Token expected");
			}
		}
		return token;
	}

}
