package org.testorrery.link;

import org.testorrery.IGenerator;
import org.testorrery.IterationException;

/**
 * Binary link that iterates only enough times to ensure that its children links have iterated completely at least once
 */
public class SequentialLink extends BinaryLink {
	
	private boolean leftCompleted = false;
	private boolean rightCompleted = false;
	
	public SequentialLink() {
		super();
	}
	
	public SequentialLink(IGenerator left, IGenerator right) {
		super(left, right);
	}
	
	public SequentialLink(IIterationLink left, IIterationLink right) {
		super(left, right);
	}
	
	/**
	 * Returns false if both branches of the link have completed at least once
	 */
	@Override
	public boolean hasNext() {
		IIterationLink left = super.getLeft();
		IIterationLink right = super.getRight();
		
		if (!left.hasNext()) {
			leftCompleted = true;
		}
		if (!right.hasNext()) {
			rightCompleted = true;
		}
		return !leftCompleted || !rightCompleted; 
	}
	
	@Override
	public void next() {
		if (!hasNext()) {
			throw new IterationException();
		}
		
		IIterationLink left = super.getLeft();
		IIterationLink right = super.getRight();
	
		if (!right.hasNext()) {
			right.reset();
		}
		if (!left.hasNext()) {
			left.reset();
		}
		
		// Note that we iterate right first because
		//  the LinkCreator usually builds trees that are shallowest
		//  to the right.  This creates less recursion overhead.
		right.next();
		left.next();		
	}
	
	@Override
	public void reset() {
		super.reset();
		leftCompleted = false;
		rightCompleted = false;
	}
	
	public void accept(IIteratorVisitor visitor) {
		visitor.visit(this);
	}
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public BinaryLink createNewInstance(IIterationLink left, IIterationLink right) {
		return new SequentialLink(left, right);
	}

}
