package org.testorrery.tests;

import org.junit.Assert;
import org.junit.Test;
import org.testorrery.Chain;
import org.testorrery.Choose;
import org.testorrery.GeneratorTester;
import org.testorrery.Repeater;
import org.testorrery.examples.IntegerRangeGenerator;

public class ChooseTest {

	@Test() 
	public void givenArray() {
		// chooses one of the values in the given list
		Choose<Integer> choose = new Choose<Integer>(1, 2, 3, 4, 5);
		GeneratorTester.testExpectedCount(choose, 1);
	}
	
	@Test()
	public void givenGenerator() {
		IntegerRangeGenerator valueGen1 = new IntegerRangeGenerator(-5, 5);
		IntegerRangeGenerator valueGen2 = new IntegerRangeGenerator(-5, 5);
		
		// Chooses one of the values in the given range
		Choose<Integer> choose1 = new Choose<Integer>(10, valueGen1);		
		GeneratorTester.testExpectedCount(choose1, 1);

		// demonstrates that Choose chooses a value randomly
		Choose<Integer> choose2 = new Choose<Integer>(10, valueGen2);		
		GeneratorTester.testExpectedCount(choose2, 1);
		
	}
	
	@Test(expected=IllegalArgumentException.class) 
	public void invalidMaxIterations() {
		Chain<Integer> valueGen = new Chain<Integer>(1);	
		// high probability that one of these will fail as expected
		new Choose<Integer>(100, valueGen);	
		new Choose<Integer>(100, valueGen);	
		new Choose<Integer>(100, valueGen);	
		new Choose<Integer>(100, valueGen);	
	}
	
	@Test() 
	public void unknownMaxIterations() {
		IntegerRangeGenerator valueGen = new IntegerRangeGenerator(1, 100);
		Assert.assertEquals(100, Choose.getMaxIterations(valueGen));		
	}
	
	@Test()
	public void changeOnReset() {
		// Tests that the setChangeOnReset accessor works.
		// Specifies whether Choose will pick a different value on reset()
		
		Choose<Integer> change = new Choose<Integer>(1, 2, 3, 4, 5);
		change.setChangeOnReset(true); // will randomly choose a new value on reset()
		change.reset();
		int changeLast = change.next();
		boolean changeDidChange = false;
		
		Choose<Integer> noChange = new Choose<Integer>(6, 7, 8, 9, 10);
		noChange.setChangeOnReset(false); // default
		noChange.reset();
		final int noChangeValue = noChange.next();		
		
		// reset a bunch of times to make sure the behavior is correct
		int count = 100;
		while (count-- > 0) {
			change.reset();
			noChange.reset();
			
			Assert.assertEquals(noChangeValue, noChange.next().intValue());
			
			if (changeLast != change.next()) {
				// may not change on every reset
				changeDidChange = true;
			}		
			System.out.println(String.format("No Change: %1$s\tShould Change: %2$s", 
					noChange.current(), 
					change.current()));
		}
		// very high probability that it changed at least once
		Assert.assertTrue(changeDidChange);		
	}
	
	
	@SuppressWarnings("unchecked")
	@Test()
	public void demoWeighting() {
		// demonstrates how one could skew the distribution of Choose
		Chain<Integer> lowValues = new Chain<Integer>(1, 2, 3);
		Chain<Integer> highValues = new Chain<Integer>(100, 200, 300);
		// low values weighted 80%; high values weighted 20%
		Chain<Integer> weightedSequence = new Chain<Integer>(
				new Repeater<Integer>(lowValues, 4),
				highValues);
		
		Choose<Integer> weighted = new Choose<Integer>(4 * 3 + 3, weightedSequence);
		weighted.setChangeOnReset(true);
		
		// demonstrates that low values are generated more often
		// than high values
		int count = 100;
		while (count-- > 0) {
			weighted.reset();
			System.out.println(weighted.next());
		}		
	}	
}
