package org.testorrery.tests.link;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.testorrery.Chain;
import org.testorrery.IGenerator;
import org.testorrery.link.CombinatoricLink;
import org.testorrery.link.GeneratorLink;
import org.testorrery.link.LinkCreator;
import org.testorrery.link.LinkExpressionParseException;
import org.testorrery.link.SequentialLink;

/**
 * Tests <code>LinkCreator</code>
 * 
 * @see LinkCreatorTest_WithWrappers
 * 
 * @author Brett Daniel
 *
 */
public class LinkCreatorTest {
	private IGenerator<String> gen0 = new Chain<String>("string1", "string2");
	private static final int gen0_COUNT = 2;
	private IGenerator<Integer> gen1 = new Chain<Integer>(1, 2, 3);
	private static final int gen1_COUNT = 3;
	private IGenerator<Character> gen2 = new Chain<Character>('a', 'b', 'c', 'd');
	private static final int gen2_COUNT = 4;
	
	@Before 
	public void setup() {
		gen0.reset();
		gen1.reset();
		gen2.reset();
	}
	
	@Test()
	public void createCombinatoricLink() {		
		CombinatoricLink link = (CombinatoricLink)LinkCreator.createCombinatoricLink(gen0, gen1, gen2);
		// check that tree was created correctly: (gen0 * gen1) * gen2
		Assert.assertEquals(gen0, ((GeneratorLink)((CombinatoricLink)link.getLeft()).getLeft()).getInnerGenerator());
		Assert.assertEquals(gen1, ((GeneratorLink)((CombinatoricLink)link.getLeft()).getRight()).getInnerGenerator());
		Assert.assertEquals(gen2, ((GeneratorLink)link.getRight()).getInnerGenerator());
		
		LinkTestUtil.iterateLink(link, gen0_COUNT * gen1_COUNT * gen2_COUNT, gen0, gen1, gen2);
	}
	
	@Test()
	public void createCombinatoricLinkIteratively() {		
		LinkCreator lc = new LinkCreator(new CombinatoricLink());
		lc.add(gen0);
		lc.add(gen1);
		lc.add(gen2);
		CombinatoricLink link = (CombinatoricLink)lc.getLink();

		// check that tree was created correctly: (gen0 * gen1) * gen2
		Assert.assertEquals(gen0, ((GeneratorLink)((CombinatoricLink)link.getLeft()).getLeft()).getInnerGenerator());
		Assert.assertEquals(gen1, ((GeneratorLink)((CombinatoricLink)link.getLeft()).getRight()).getInnerGenerator());
		Assert.assertEquals(gen2, ((GeneratorLink)link.getRight()).getInnerGenerator());
		
		LinkTestUtil.iterateLink(link, gen0_COUNT * gen1_COUNT * gen2_COUNT, gen0, gen1, gen2);
	}
	
	@Test()
	public void createSequentialLink() {		
		SequentialLink link = (SequentialLink)LinkCreator.createSequentialLink(gen0, gen1, gen2);		
		// check that tree was created correctly: (gen0 + gen1) + gen2
		Assert.assertEquals(gen0, ((GeneratorLink)((SequentialLink)link.getLeft()).getLeft()).getInnerGenerator());
		Assert.assertEquals(gen1, ((GeneratorLink)((SequentialLink)link.getLeft()).getRight()).getInnerGenerator());
		Assert.assertEquals(gen2, ((GeneratorLink)link.getRight()).getInnerGenerator());
		
		LinkTestUtil.iterateLink(link, gen2_COUNT, gen0, gen1, gen2);
	}
	
	@Test()
	public void createSequentialLinkIteratively() {		
		LinkCreator lc = new LinkCreator(new SequentialLink());
		lc.add(gen0);
		lc.add(gen1);
		lc.add(gen2);
		SequentialLink link = (SequentialLink)lc.getLink();

		// check that tree was created correctly: (gen0 + gen1) + gen2
		Assert.assertEquals(gen0, ((GeneratorLink)((SequentialLink)link.getLeft()).getLeft()).getInnerGenerator());
		Assert.assertEquals(gen1, ((GeneratorLink)((SequentialLink)link.getLeft()).getRight()).getInnerGenerator());
		Assert.assertEquals(gen2, ((GeneratorLink)link.getRight()).getInnerGenerator());
		
		LinkTestUtil.iterateLink(link, gen2_COUNT, gen0, gen1, gen2);
	}
	
	@Test()
	public void validExpression() {		
		testExpression("($0 + $1) * $2", gen1_COUNT * gen2_COUNT, gen0, gen1, gen2);
	}
	
	@Test()
	public void whitespace() {
		testExpression("   (   $\n\n0\r\f\t  +  $ 1 ) * $ 2", gen1_COUNT * gen2_COUNT, gen0, gen1, gen2);
	}
	
	@Test(expected=LinkExpressionParseException.class)
	public void unbalancedParentheses1() {
		testExpression("$0 + $1) * $2", -1, gen0, gen1, gen2);
	}
	
	@Test(expected=LinkExpressionParseException.class)
	public void unbalancedParentheses2() {
		testExpression("($0 + $1 * $2", -1, gen0, gen1, gen2);
	}
	
	@Test()
	public void nestedParentheses() {
		testExpression("(((((((($0 * $1) * $2)))))))", gen0_COUNT * gen1_COUNT * gen2_COUNT, gen0, gen1, gen2);
	}
	
	@Test(expected=LinkExpressionParseException.class) 
	public void invalidOperation() {
		testExpression("$0 - $1", -1, gen0, gen1);
	}
	
	@Test(expected=LinkExpressionParseException.class) 
	public void invalidGeneratorIndex() {
		testExpression("$5 * $0", -1, gen0, gen1);
	}
	
	@Test(expected=LinkExpressionParseException.class) 
	public void duplicateGeneratorIndex() {
		testExpression("$0 * $0", -1, gen0, gen1);
	}
	
	private static void testExpression(String expression, int expectedCount, IGenerator... generators) {
		LinkTestUtil.iterateLink(
				LinkCreator.createLinkFromExpression(expression, generators),
				expectedCount, 
				generators);
	}
	
}
