package edu.uiuc.astgen.basic;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.jdt.core.dom.BodyDeclaration;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.SimpleType;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.testorrery.IGenerator;
import org.testorrery.link.CombinatoricLink;
import org.testorrery.link.IIterationLink;
import org.testorrery.link.LinkCreator;

import edu.uiuc.astgen.ASTNodeGeneratorBase;

/**
 * TypeDeclaration generator that returns class declarations
 */
public class ClassGenerator extends ASTNodeGeneratorBase<TypeDeclaration> {	
	
	private List<IGenerator<? extends BodyDeclaration>> bodyDeclarationGenerators;
	private IGenerator<SimpleName> identifierGen;
	private IGenerator<Modifier> visibilityModifierGen;
	private IGenerator<Modifier> classModifierGen;
	
	private Type parentClass;	
	private List<SimpleType> listInterfaces;
	private boolean isAnInternalClass; 
	
	public ClassGenerator() {
		super();		
		bodyDeclarationGenerators = new LinkedList<IGenerator<? extends BodyDeclaration>>();
		listInterfaces = new ArrayList<SimpleType>();
	}
	
	public ClassGenerator(String identifier) {
		this(new SimpleNameGenerator(identifier));
	}	
	
	public ClassGenerator(IGenerator<SimpleName> simpleNameGenerator) {
		this();
		identifierGen = simpleNameGenerator;
	}
	
	@Override
	protected IIterationLink getInitializedIterator() {
		if (identifierGen == null) {
			identifierGen = new SimpleNameGenerator("TheClass");
		}
		if (visibilityModifierGen == null) {
			visibilityModifierGen = new ModifierGenerator((Modifier.ModifierKeyword)null); // absence of a visibility modifier
		}
		if (classModifierGen == null) {
			classModifierGen = new ModifierGenerator((Modifier.ModifierKeyword)null); // absence of a class modifier
		}
		
		LinkCreator lc = new LinkCreator(new CombinatoricLink());
		lc.add(identifierGen);
		lc.add(visibilityModifierGen);
		lc.add(classModifierGen);
		for (IGenerator<? extends BodyDeclaration> bodyDeclGen : bodyDeclarationGenerators) {
			lc.add(bodyDeclGen);
		}
		return lc.getLink();		
	}

	public IGenerator<Modifier> getClassModifierGen() {
		return classModifierGen;
	}

	public void setParentClass(Type parent) {
		parentClass = parent;
	}
	
	public Type getParentClass () {
		return parentClass;
	}
	
	public void setClassModifierGen(IGenerator<Modifier> modifierGen) {
		classModifierGen = modifierGen;
	}

	public IGenerator<SimpleName> getIdentifierGen() {
		return identifierGen;
	}

	public void setIdentifierGen(IGenerator<SimpleName> gen) {
		identifierGen = gen;
	}

	public IGenerator<Modifier> getVisibilityModifierGen() {
		return visibilityModifierGen;
	}

	public void setVisibilityModifierGen(IGenerator<Modifier> modifierGen) {
		visibilityModifierGen = modifierGen;
	}

	public void addBodyDeclarationGenerator(IGenerator<? extends BodyDeclaration> bodyDeclarationGenerator) {
		bodyDeclarationGenerators.add(bodyDeclarationGenerator);
	}
	
	public void setBodyDeclarationGenerator(List<IGenerator<? extends BodyDeclaration>> listBodysDeclarations) {
		bodyDeclarationGenerators = listBodysDeclarations;
	}

	public List<IGenerator<? extends BodyDeclaration>> getBodyDeclarationGenerator() {
		return bodyDeclarationGenerators;
	}
	
	
	public void addInterface(SimpleType interfaceDeclaration) {
		listInterfaces.add(interfaceDeclaration);
	}
	
	public void setListOfInterfaces (List<SimpleType> newListInterfaces) {
		listInterfaces = newListInterfaces;
	}
	
	@Override
	protected TypeDeclaration generateCurrent() {		
		TypeDeclaration toReturn = super.getAST().newTypeDeclaration();
		toReturn.setInterface(false);
		if (parentClass != null) {
			if (parentClass.getParent() != null || parentClass.getAST() != super.getAST())
				parentClass = (SimpleType)parentClass.copySubtree(super.getAST(), parentClass);
			try {	
				toReturn.setSuperclassType(parentClass);
				} catch (Exception e) {
					System.out.println("error setting parentClass: "+ e);
				}			
		}
		for (int i = 0; i < listInterfaces.size(); i++) {
			SimpleType interfaceDeclaration=listInterfaces.get(i);
			if (interfaceDeclaration.getParent() != null)
				interfaceDeclaration = (SimpleType)interfaceDeclaration.copySubtree(super.getAST(), interfaceDeclaration);
			toReturn.superInterfaceTypes().add(interfaceDeclaration);			
		}
		
		toReturn.setName(identifierGen.current());
		
		Modifier visibilityModifier = visibilityModifierGen.current();
		if (visibilityModifier != null) {
			toReturn.modifiers().add(visibilityModifier);
		}
		
		Modifier classModifier = classModifierGen.current();
		if (classModifier != null) {
			toReturn.modifiers().add(classModifier);	
		}
		
		for (IGenerator<? extends BodyDeclaration> bodyDeclGen : bodyDeclarationGenerators) {
			toReturn.bodyDeclarations().add(bodyDeclGen.current());
		}
		return toReturn;
	}
	
	// List of default constructors
	public static ClassGenerator getNewDefaultClassGenerator(String name) {
		ClassGenerator defaultClassGen = new ClassGenerator();
		// default generators 
		defaultClassGen.setIdentifierGen(new SimpleNameGenerator(name));
		defaultClassGen.setVisibilityModifierGen(new ModifierGenerator((Modifier.ModifierKeyword)null));
		defaultClassGen.setClassModifierGen(new ModifierGenerator((Modifier.ModifierKeyword)null)); // absence of a class modifier
		defaultClassGen.setListOfInterfaces(new ArrayList<SimpleType>());
		defaultClassGen.setBodyDeclarationGenerator(new LinkedList<IGenerator<? extends BodyDeclaration>>());
		return defaultClassGen;
	}

	public boolean isAnInternalClass() {
		return isAnInternalClass;
	}

	public void setAnInternalClass(boolean isAnInternalClass) {
		this.isAnInternalClass = isAnInternalClass;
	}
}
