package edu.uiuc.astgen.basic;

import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.PrimitiveType;
import org.eclipse.jdt.core.dom.SimpleType;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.testorrery.IGenerator;
import org.testorrery.link.IIterationLink;
import org.testorrery.link.LinkCreator;

import edu.uiuc.astgen.ASTNodeGeneratorBase;
import edu.uiuc.astgen.ASTNodeLiteral;

/**
 * Field declaration generator
 */
public class FieldDeclarationGenerator extends ASTNodeGeneratorBase<FieldDeclaration> {
	private IGenerator<Modifier> visibilityModifierGen; // public, private, etc.
	private IGenerator<Modifier> fieldModifierGen; // static, final, etc.
	private IGenerator<? extends Type> typeGen; 
	private IGenerator<VariableDeclarationFragment> variableDeclarationFragmentGen; // identifier 
		
	public FieldDeclarationGenerator(String name) {
		this(new VariableDeclarationFragmentGenerator(new SimpleNameGenerator(name)));
	}
	
	public FieldDeclarationGenerator(
			IGenerator<VariableDeclarationFragment> variableDeclarationFragmentGen) {
		super();
		
		this.typeGen = new ASTNodeLiteral<SimpleType>(super.getAST().newSimpleType(super.getAST().newName("Object")));
		this.visibilityModifierGen = new ModifierGenerator(Modifier.ModifierKeyword.PUBLIC_KEYWORD);
		this.fieldModifierGen = new ModifierGenerator((Modifier.ModifierKeyword)null); // absent field modifier
		this.variableDeclarationFragmentGen = variableDeclarationFragmentGen;
		
		// _mainIterator initialized on first reference
	}
	
	public FieldDeclarationGenerator(
			IGenerator<Modifier> visibilityGen,
			IGenerator<Modifier> fieldModifierGen,
			IGenerator<? extends Type> primitiveTypeGen,
			IGenerator<VariableDeclarationFragment> variableDeclarationFragmentGen) {
		super();
		
		this.typeGen = primitiveTypeGen;
		this.visibilityModifierGen = visibilityGen;
		this.fieldModifierGen = fieldModifierGen;
		this.variableDeclarationFragmentGen = variableDeclarationFragmentGen;
		
	}
	
	@Override
	protected IIterationLink getInitializedIterator() {
		return LinkCreator.createCombinatoricLink(
				typeGen,
				visibilityModifierGen,
				fieldModifierGen,
				variableDeclarationFragmentGen);
	}	
	
	public IGenerator<Modifier> getFieldModifierGen() {
		return fieldModifierGen;
	}

	public void setFieldModifierGen(IGenerator<Modifier> modifierGen) {
		fieldModifierGen = modifierGen;
	}

	public IGenerator<? extends Type> getTypeGen() {
		return typeGen;
	}

	public void setTypeGen(IGenerator<? extends Type> typeGen) {
		this.typeGen = typeGen;
	}

	public IGenerator<VariableDeclarationFragment> getVariableDeclarationFragmentGen() {
		return variableDeclarationFragmentGen;
	}

	public void setVariableDeclarationFragmentGen(
			IGenerator<VariableDeclarationFragment> declarationFragmentGen) {
		variableDeclarationFragmentGen = declarationFragmentGen;
	}

	public IGenerator<Modifier> getVisibilityModifierGen() {
		return visibilityModifierGen;
	}

	public void setVisibilityModifierGen(IGenerator<Modifier> modifierGen) {
		visibilityModifierGen = modifierGen;
	}

	@Override
	protected FieldDeclaration generateCurrent() {
		VariableDeclarationFragment vdf = variableDeclarationFragmentGen.current();
		Type fieldType = typeGen.current();
		Modifier visibilityModifier = visibilityModifierGen.current();
		Modifier fieldModifier = fieldModifierGen.current();		

		FieldDeclaration fd = super.getAST().newFieldDeclaration(vdf);
		fd.setType(fieldType);
		if (visibilityModifier != null) {
			fd.modifiers().add(visibilityModifier);
		}
		if (fieldModifier != null) {
			fd.modifiers().add(fieldModifier);	
		}			
		
		return fd;
	}

	// List of default constructors
	public static FieldDeclarationGenerator getNewDefaultFieldDeclarationGenerator(String name) {
		PrimitiveTypeGenerator primitiveTypeGenerator = new PrimitiveTypeGenerator(
				PrimitiveType.BOOLEAN, PrimitiveType.CHAR, PrimitiveType.INT);
		VariableDeclarationFragmentGenerator variableDeclarationFragmentGenerator = new VariableDeclarationFragmentGenerator(
				new SimpleNameGenerator(name));
		FieldDeclarationGenerator fGenerator = new FieldDeclarationGenerator(
				ModifierGenerator.createVisibilityModifier(), ModifierGenerator.createFieldModifier(),
				primitiveTypeGenerator,
				variableDeclarationFragmentGenerator);		
		return fGenerator;
	}
}
