package edu.uiuc.astgen.basic;

import java.util.ArrayList;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.PrimitiveType;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.Type;
import org.testorrery.IGenerator;
import org.testorrery.link.CombinatoricLink;
import org.testorrery.link.IIterationLink;
import org.testorrery.link.LinkCreator;

import edu.uiuc.astgen.ASTNodeChain;
import edu.uiuc.astgen.ASTNodeGeneratorBase;
import edu.uiuc.astgen.ASTNodeLiteral;

/**
 * Generates MethodDeclaration objects
 */
public class MethodDeclarationGenerator extends
		ASTNodeGeneratorBase<MethodDeclaration> {
	
	private IGenerator<SimpleName> _nameGen;
	private IGenerator<? extends Type> _returnTypeGen;	
	private IGenerator<Modifier> _visibilityModifierGen;
	private IGenerator<Modifier> _methodModifierGen;
	private IGenerator<Block> _methodBodyGen;
	private ArrayList<IGenerator<SingleVariableDeclaration>> _parameterGens;
	
	public MethodDeclarationGenerator(String identifier) {
		this(new SimpleNameGenerator(identifier));
	}
	
	public MethodDeclarationGenerator(IGenerator<SimpleName> nameGen) {	
		super();		
		_nameGen = nameGen;
		_returnTypeGen = new PrimitiveTypeGenerator(PrimitiveType.VOID);
		_visibilityModifierGen = new ModifierGenerator(Modifier.ModifierKeyword.PUBLIC_KEYWORD);
		_methodModifierGen = new ModifierGenerator((Modifier.ModifierKeyword)null); // absence of a method modifier
		_methodBodyGen = new ASTNodeLiteral<Block>(super.getAST().newBlock()); // empty block;
		_parameterGens = new ArrayList<IGenerator<SingleVariableDeclaration>>(); // empty parameter list
	}

	@Override
	protected IIterationLink getInitializedIterator() {
		LinkCreator linkCreator = new LinkCreator(new CombinatoricLink());
		linkCreator.add(_nameGen);
		linkCreator.add(_returnTypeGen);
		linkCreator.add(_visibilityModifierGen);
		linkCreator.add(_methodModifierGen);
		linkCreator.add(_methodBodyGen);
		linkCreator.addAll(_parameterGens);
		return linkCreator.getLink();
	}
	
	public void setNameGenerator(IGenerator<SimpleName> nameGen) {
		_nameGen = nameGen;
	}	
	
	public void setMethodBodyGenerator(IGenerator<Block> bodyGen) {
		_methodBodyGen = bodyGen;
	}

	public void setMethodModifierGenerator(IGenerator<Modifier> modifierGen) {
		_methodModifierGen = modifierGen;
	}

	public void setReturnTypeGenerator(IGenerator<? extends Type> typeGen) {
		_returnTypeGen = typeGen;
	}

	public void setVisibilityModifierGenerator(IGenerator<Modifier> modifierGen) {
		_visibilityModifierGen = modifierGen;
	}
	
	public void addParameterGenerator(IGenerator<SingleVariableDeclaration> parameterGen) {
		_parameterGens.add(parameterGen);
	}

	@Override
	protected MethodDeclaration generateCurrent() {
		MethodDeclaration method = super.getAST().newMethodDeclaration();
		method.setName(_nameGen.current());
		method.setReturnType2(_returnTypeGen.current());	
		Modifier visibilityModifier = _visibilityModifierGen.current();
		if (visibilityModifier != null) {
			method.modifiers().add(visibilityModifier);
		}
		Modifier methodModifier = _methodModifierGen.current();
		if (methodModifier != null) {
			method.modifiers().add(methodModifier);
		}		
		method.setBody(_methodBodyGen.current());
		for (IGenerator<SingleVariableDeclaration> paramGen : _parameterGens) {
			SingleVariableDeclaration param = paramGen.current();
			if (param != null) {
				method.parameters().add(param);
			}
		}		
		
		return method;
	}
	
	// List of default constructors
	public static MethodDeclarationGenerator getNewDefaultMethodDeclarationGenerator (String name) {
		AST ast = ast();
		SimpleNameGenerator nameGen = new SimpleNameGenerator(name);
		PrimitiveTypeGenerator returnTypeGen = new PrimitiveTypeGenerator(PrimitiveType.INT,
				PrimitiveType.LONG);
		ModifierGenerator visibilityModifierGen = new ModifierGenerator(Modifier.ModifierKeyword.PUBLIC_KEYWORD,
				Modifier.ModifierKeyword.PRIVATE_KEYWORD);
		ModifierGenerator methodModifierGen = new ModifierGenerator(Modifier.ModifierKeyword.ABSTRACT_KEYWORD,
				Modifier.ModifierKeyword.FINAL_KEYWORD);
		IGenerator<Block> methodBodyGen = new ASTNodeLiteral<Block>(
				ast.newBlock()); // TODO: non-empty
		// block
		ASTNodeChain<SingleVariableDeclaration> paramSequence = new ASTNodeChain<SingleVariableDeclaration>(
				new SingleVariableDeclarationGenerator(new SimpleNameGenerator("intParam"),
						new PrimitiveTypeGenerator(PrimitiveType.INT)), new SingleVariableDeclarationGenerator(new SimpleNameGenerator("longParam"),
						new PrimitiveTypeGenerator(PrimitiveType.LONG))); // chooses long or int as parameter 

		MethodDeclarationGenerator methodGen = new MethodDeclarationGenerator(
				nameGen);
		methodGen.setReturnTypeGenerator(returnTypeGen);
		methodGen.setVisibilityModifierGenerator(visibilityModifierGen);
		methodGen.setMethodModifierGenerator(methodModifierGen);
		methodGen.setMethodBodyGenerator(methodBodyGen);
		methodGen.addParameterGenerator(paramSequence);
		
		return methodGen;
	}	
}
