package edu.uiuc.astgen.classrelationships;

import java.util.LinkedList;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.core.dom.VariableDeclarationExpression;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;

import edu.uiuc.astgen.ASTNodeGeneratorBase;

/**
 * Utility class used in various {@link RelationshipBuilder}s to build abstract syntax trees
 */
public class BuilderUtil {
	private static final String CONTAINER_NAME_PLACEHOLDER = "<CONTAINER>"; 
		
	private static final String DEFAULT_INNER_CLASS_NAME = "DummyClass_<CONTAINER>";
	private static final String DEFAULT_METHOD_NAME = "dummyMethod_<CONTAINER>";
	private static final String DEFAULT_FIELD_NAME = "dummyField_<CONTAINER>";
	private static final String DEFAULT_VARIABLE_NAME = "dummyVariable_<CONTAINER>";
	
	public static String getDummyInnerClassName(String containerName) {
		return DEFAULT_INNER_CLASS_NAME.replace(CONTAINER_NAME_PLACEHOLDER, containerName);
	}
	public static String getDummyFieldName(String containerName) {
		return DEFAULT_FIELD_NAME.replace(CONTAINER_NAME_PLACEHOLDER, containerName);
	}
	public static String getDummyMethodName(String containerName) {
		return DEFAULT_METHOD_NAME.replace(CONTAINER_NAME_PLACEHOLDER, containerName);
	}
	public static String getDummyVariableName(String containerName) {
		return DEFAULT_VARIABLE_NAME.replace(CONTAINER_NAME_PLACEHOLDER, containerName);
	}	
	
	/**
	 * Builds a unique {@link FieldDeclaration} for the given {@link TypeDeclaration}.
	 * Note: created field is not added to type declaration. 
	 */
	public static FieldDeclaration getDummyField(TypeDeclaration container) {
		return getDummyField(container, null);
	}	
	
	/**
	 * Builds a unique {@link FieldDeclaration} for the given {@link TypeDeclaration}.
	 * Note: created field is not added to type declaration. 
	 */
	public static FieldDeclaration getDummyField(TypeDeclaration container, Expression initializer) {
		// build existing field name list
		LinkedList<String> existingFields = new LinkedList<String>();
		for (Object declaration : container.bodyDeclarations()) {
			if (declaration instanceof FieldDeclaration) {
				for (Object fragment : ((FieldDeclaration)declaration).fragments()) {
					existingFields.add(((VariableDeclarationFragment)fragment).getName().toString());
				}
			}
		}
		String newFieldName = createNextNumberedIdentifier(existingFields, DEFAULT_FIELD_NAME, container.getName().toString());
		
		// build FieldDeclaration
		AST ast = ASTNodeGeneratorBase.ast();
		VariableDeclarationFragment varDeclFrag = ast.newVariableDeclarationFragment();
		varDeclFrag.setName(ast.newSimpleName(newFieldName));
		varDeclFrag.setInitializer(initializer);
		FieldDeclaration fieldDecl = ast.newFieldDeclaration(varDeclFrag);
		fieldDecl.modifiers().add(ast.newModifier(Modifier.ModifierKeyword.PUBLIC_KEYWORD));
		
		return fieldDecl;
	}
	
	/**
	 * Builds a unique {@link MethodDeclaration} for the given {@link TypeDeclaration}.
	 * Note: method is not added to type declaration. Caller must set body, return type, parameters, etc. 
	 */
	public static MethodDeclaration getDummyMethod(TypeDeclaration container) {
		// build existing method name list
		LinkedList<String> existingMethods = new LinkedList<String>();
		for (Object declaration : container.bodyDeclarations()) {
			if (declaration instanceof MethodDeclaration) {
				existingMethods.add(((MethodDeclaration)declaration).getName().toString());
			}
		}
		String newMethodName = createNextNumberedIdentifier(existingMethods, DEFAULT_METHOD_NAME, container.getName().toString());
		
		// build MethodDeclaration
		AST ast = ASTNodeGeneratorBase.ast();
		MethodDeclaration methDecl = ast.newMethodDeclaration();
		methDecl.setName(ast.newSimpleName(newMethodName));
		methDecl.modifiers().add(ast.newModifier(Modifier.ModifierKeyword.PUBLIC_KEYWORD));
		methDecl.setBody(ast.newBlock());
		
		return methDecl;
	}
	
	/**
	 * Returns the first dummy method found or if none exist, creates and returns a new dummy method.
	 * Note: if no method is found, the new dummy method will have been added to the type declaration.
	 */
	public static MethodDeclaration getExistingDummyMethod(TypeDeclaration container) {
		String expectedName = DEFAULT_METHOD_NAME.replace(CONTAINER_NAME_PLACEHOLDER, container.getName().toString());
		
		// search for existing dummy method
		for (Object declaration : container.bodyDeclarations()) {
			if (declaration instanceof MethodDeclaration && ((MethodDeclaration)declaration).getName().toString().startsWith(expectedName)) {
				return (MethodDeclaration)declaration;
			}
		}
		
		// create a method if one does not exist
		MethodDeclaration methDecl = getDummyMethod(container);
		container.bodyDeclarations().add(methDecl);
		return methDecl;
	}
	
	/**
	 * Builds a unique {@link VariableDeclarationExpression} for the given method.
	 */
	public static VariableDeclarationExpression getDummyVariable(MethodDeclaration methDecl) {
		// build list of existing variables
		LinkedList<String> existing = new LinkedList<String>();
		for (Object param : methDecl.parameters()) {
			existing.add(((SingleVariableDeclaration)param).getName().toString());
		}
		for (Object stmt : methDecl.getBody().statements()) {
			if (stmt instanceof VariableDeclarationExpression) {
				for (Object fragment : ((VariableDeclarationExpression)stmt).fragments()) {
					existing.add(((VariableDeclarationFragment)fragment).getName().toString());
				}
			}
		}
		String newName = createNextNumberedIdentifier(existing, DEFAULT_VARIABLE_NAME, methDecl.getName().toString());
		
		AST ast = ASTNodeGeneratorBase.ast();
		VariableDeclarationFragment frag = ast.newVariableDeclarationFragment();
		frag.setName(ast.newSimpleName(newName));
		VariableDeclarationExpression varDeclExpr = ast.newVariableDeclarationExpression(frag);
		varDeclExpr.setType(ast.newSimpleType(ast.newSimpleName("Object")));
		return varDeclExpr;
	}
	
	/**
	 * Creates a unique identifier  
	 * @param existingNames The identifiers that already exist in the container
	 * @param defaultName The default identifier (a number will be appended to the default to ensure uniqueness)
	 * @param containerName The name of the container that will hold the identifier
	 */
	private static String createNextNumberedIdentifier(Iterable<String> existingNames, String defaultName, String containerName) {
		
		// expand container placeholder in default name
		defaultName = defaultName.replace(CONTAINER_NAME_PLACEHOLDER, containerName);
		
		int newSuffix = 0;
		int curSuffix = 0;		
		for (String curName: existingNames) {
			if (curName.startsWith(defaultName)) {
				if (newSuffix == 0 && curName.length() == defaultName.length()) {
					// found the default name exactly
					newSuffix = 1;
				}
				else {
					// found the default name with some suffix
					try {
						curSuffix = Integer.parseInt(curName.substring(defaultName.length()));
						if (curSuffix >= newSuffix) {
							// new method suffix should come after largest existing suffix
							newSuffix = curSuffix + 1;
						}
					}
					catch (NumberFormatException e) {
						// do nothing. suffix was not one of ours
					}
				}
			}
		}
		
		if (newSuffix == 0) {
			return defaultName;
		}
		return defaultName + newSuffix;
	}

}
