package edu.uiuc.astgen.classrelationships;

import org.eclipse.jdt.core.dom.*;

import edu.uiuc.astgen.ASTNodeGeneratorBase;

/**
 * Enum of ways in which a particular class may be referenced in another class.
 */
public enum ReferenceType implements RelationshipSpecification {

	NO_REFERENCE(new NoReferenceBuilder(), RelationshipDirection.UNRELATED, Reversibility.NOT_REVERSIBLE),
	
	/// Declaration reference types
	
	/**
	 * class B {
	 *   B(A arg0) {}
	 * }
	 */
	CONSTRUCTOR_PARAMETER_TYPE(new ConstructorParameterTypeBuilder(), RelationshipDirection.B_TO_A, Reversibility.IS_REVERSIBLE),
	/**
	 * class B {
	 *   A dummyField;
	 * } 
	 */
	FIELD_TYPE(new FieldTypeBuilder(), RelationshipDirection.B_TO_A, Reversibility.IS_REVERSIBLE),
	/**
	 * class B {
	 *   void dummyMethod(A arg0) {}
	 * }
	 */
	PARAMETER_TYPE(new ParameterTypeBuilder(), RelationshipDirection.B_TO_A, Reversibility.IS_REVERSIBLE),
	/**
	 * class B {
	 *   A dummyMethod() {}
	 * }
	 */
	RETURN_TYPE(new ReturnTypeBuilder(), RelationshipDirection.B_TO_A, Reversibility.IS_REVERSIBLE),
	/**
	 * class B {
	 *   void dummyMethod() throws A {}
	 * }
	 */
	THROWS_CLAUSE_TYPE(new ThrowsClauseTypeBuilder(), RelationshipDirection.B_TO_A, Reversibility.IS_REVERSIBLE),
	
	/// Statement reference types
	
	/**
	 * try {}
	 * catch (A e) {} 
	 */
	CATCH_CLAUSE_PARAMETER_TYPE(new CatchClauseParameterTypeBuilder(), RelationshipDirection.B_TO_A, Reversibility.IS_REVERSIBLE),
	/**
	 * A dummyVariable;
	 */
	LOCAL_VARIABLE_TYPE(new LocalVariableTypeBuilder(), RelationshipDirection.B_TO_A, Reversibility.IS_REVERSIBLE),
	/**
	 * for (A i;;) {}
	 */
	FOR_LOOP_VARIABLE_TYPE(new ForLoopVariableTypeBuilder(), RelationshipDirection.B_TO_A, Reversibility.IS_REVERSIBLE),
	/**
	 * for (A element : elements) {}
	 */
	FOREACH_LOOP_VARIABLE_TYPE(new ForeachLoopVariableTypeBuilder(), RelationshipDirection.B_TO_A, Reversibility.IS_REVERSIBLE),
	/**
	 * A.dummyMethod()
	 */
	STATIC_METHOD_CALL(new StaticMethodCallBuilder(), RelationshipDirection.B_TO_A, Reversibility.IS_REVERSIBLE),
	
	/// Expression reference types
	
	/**
	 * new A(...)
	 */
	CONSTRUCTOR_CALL_EXPRESSION(new ConstructorCallBuilder(), RelationshipDirection.B_TO_A, Reversibility.IS_REVERSIBLE),
	/**
	 * new A[] {}
	 */
	ARRAY_CREATION_EXPRESSION(new ArrayCreationBuilder(), RelationshipDirection.B_TO_A, Reversibility.IS_REVERSIBLE),
	/**
	 * (A)null
	 */
	CAST_EXPRESSION(new CastBuilder(), RelationshipDirection.B_TO_A, Reversibility.IS_REVERSIBLE),
	/**
	 * null instanceof A
	 */
	INSTANCEOF_EXPRESSION(new InstanceofBuilder(), RelationshipDirection.B_TO_A, Reversibility.IS_REVERSIBLE),
	/**
	 * A.class
	 */
	TYPE_LITERAL_EXPRESSION(new TypeLiteralBuilder(), RelationshipDirection.B_TO_A, Reversibility.IS_REVERSIBLE),
	/**
	 * A.dummyField
	 */
	STATIC_FIELD_REFERENCE_EXPRESSION(new StaticFieldReferenceBuilder(), RelationshipDirection.B_TO_A, Reversibility.IS_REVERSIBLE),
	/**
	 * A.this
	 */
	QUALIFIED_THIS_EXPRESSION(new QualifiedThisBuilder(), RelationshipDirection.B_TO_A, Reversibility.IS_REVERSIBLE);

	private static AST ast = ASTNodeGeneratorBase.ast();
	
	/**
	 * Default builder that does nothing
	 */
	private static class NoReferenceBuilder implements RelationshipBuilder {
		public void build(TypeDeclaration a, TypeDeclaration b) {
			// do nothing
		}
	}
	
	
	
	//// Declaration reference builders
	
	/**
	 * Base class for builders that build some kind of class-level declaration
	 */
	private static abstract class DeclarationReferenceBuilder implements RelationshipBuilder {
		public void build(TypeDeclaration a, TypeDeclaration b) {			
			BodyDeclaration bodyDecl = getDeclaration(a, b);
			b.bodyDeclarations().add(bodyDecl);			
		}
		
		protected abstract BodyDeclaration getDeclaration(TypeDeclaration a, TypeDeclaration b);
	}
		
	private static class ConstructorParameterTypeBuilder extends DeclarationReferenceBuilder {	
		@Override
		protected BodyDeclaration getDeclaration(TypeDeclaration a, TypeDeclaration b) {			
			/* 
			 * B(A arg0) {}
			 */			
			
			// build method parameter
			SingleVariableDeclaration paramDecl = ast.newSingleVariableDeclaration();
			paramDecl.setName(ast.newSimpleName("arg0"));
			paramDecl.setType(ast.newSimpleType(ast.newSimpleName(a.getName().toString())));			
			
			// build constructor
			Block methBody = ast.newBlock();			
			MethodDeclaration methDecl = ast.newMethodDeclaration();
			methDecl.setConstructor(true);			
			methDecl.modifiers().add(ast.newModifier(Modifier.ModifierKeyword.PUBLIC_KEYWORD));
			methDecl.setName(ast.newSimpleName(b.getName().toString()));
			methDecl.setBody(methBody);
			methDecl.parameters().add(paramDecl);
			
			return methDecl;
		}
	}
	
	private static class FieldTypeBuilder extends DeclarationReferenceBuilder {		
		@Override
		protected BodyDeclaration getDeclaration(TypeDeclaration a, TypeDeclaration b) {	
			/*
			 * B dummyField;
			 */
			
			// build field
			FieldDeclaration fieldDecl = BuilderUtil.getDummyField(b);
			// field's type is the target type
			fieldDecl.setType(ast.newSimpleType((Name)ASTNode.copySubtree(ast, a.getName())));
			
			return fieldDecl;
		}
	}
	
	private static class ParameterTypeBuilder extends DeclarationReferenceBuilder {
		@Override
		protected BodyDeclaration getDeclaration(TypeDeclaration a, TypeDeclaration b) {				
			/* 
			 * void dummyMethod(A arg0) {}
			 */
			
			// build method parameter
			SingleVariableDeclaration paramDecl = ast.newSingleVariableDeclaration();
			paramDecl.setName(ast.newSimpleName("arg0"));
			paramDecl.setType(ast.newSimpleType(ast.newSimpleName(a.getName().toString())));			
			
			// build method
			Block methBody = ast.newBlock();
			MethodDeclaration methDecl = BuilderUtil.getDummyMethod(b);
			methDecl.setBody(methBody);
			methDecl.parameters().add(paramDecl);
			
			return methDecl;
		}
	}
	
	private static class ReturnTypeBuilder extends DeclarationReferenceBuilder {
		@Override
		protected BodyDeclaration getDeclaration(TypeDeclaration a, TypeDeclaration b) {
			/* 
			 * A dummyMethod() {
			 *   return null;
			 * }
			 */
			
			// build return statement			
			ReturnStatement retStmt = ast.newReturnStatement();
			retStmt.setExpression(ast.newNullLiteral());
			
			// build method			
			Block methBody = ast.newBlock();
			methBody.statements().add(retStmt);
			MethodDeclaration methDecl = BuilderUtil.getDummyMethod(b);
			methDecl.setBody(methBody);
			methDecl.setReturnType2(ast.newSimpleType(ast.newSimpleName(a.getName().toString())));
			
			return methDecl;
		}
	}
	
	private static class ThrowsClauseTypeBuilder extends DeclarationReferenceBuilder {
		@Override
		protected BodyDeclaration getDeclaration(TypeDeclaration a, TypeDeclaration b) {
			/* 
			 * void dummyMethod() throws A {}
			 */
			
			// build method
			Block methBody = ast.newBlock();
			MethodDeclaration methDecl = BuilderUtil.getDummyMethod(b);
			methDecl.setBody(methBody);
			methDecl.thrownExceptions().add(ast.newSimpleName(a.getName().toString()));

			return methDecl;
		}
	}
	
	
	
	//// Statement Reference Builders 
	
	/**
	 * Base class for builders that create a method-level statement
	 */
	private abstract static class StatementReferenceBuilder implements RelationshipBuilder {
		public void build(TypeDeclaration a, TypeDeclaration b) {			
			// find the target method in b
			MethodDeclaration dummyMethod = BuilderUtil.getExistingDummyMethod(b);
			
			// add the statement to the method
			dummyMethod.getBody().statements().add(getStatement(a));
		}
		protected abstract Statement getStatement(TypeDeclaration targetType);
	}
	
	private static class CatchClauseParameterTypeBuilder extends StatementReferenceBuilder {
		@Override
		protected Statement getStatement(TypeDeclaration targetType) {			
			/* 
			 * try {}
			 * catch (A e) {}
			 */

			// build tryStatement
			CatchClause catchClause = ast.newCatchClause();
			SingleVariableDeclaration svd = ast.newSingleVariableDeclaration();
			svd.setName(ast.newSimpleName("e"));
			svd.setType(ast.newSimpleType(ast.newSimpleName(targetType.getName().toString())));
			catchClause.setException(svd);
			TryStatement tryStatement = ast.newTryStatement();
			tryStatement.catchClauses().add(catchClause);
			
			return tryStatement;
		}
	}
	
	private static class LocalVariableTypeBuilder extends StatementReferenceBuilder {
		@Override
		protected Statement getStatement(TypeDeclaration targetType) {
			/*
			 * A theA;
			 */
			
			// build local variable declaration
			VariableDeclarationFragment varDeclFrag = ast.newVariableDeclarationFragment();
			varDeclFrag.setName(ast.newSimpleName("the" + targetType.getName().toString()));
			VariableDeclarationExpression varDeclExpr = ast.newVariableDeclarationExpression(varDeclFrag);
			// variable's type is A
			varDeclExpr.setType(ast.newSimpleType(ast.newSimpleName(targetType.getName().toString())));
			
			return ast.newExpressionStatement(varDeclExpr);
		}
	}
	
	private static class ForLoopVariableTypeBuilder extends StatementReferenceBuilder {
		@Override
		protected Statement getStatement(TypeDeclaration targetType) {
			/* 
			 * for (A i;;) {}
			 */
			
			// build forStatement
			ForStatement forStatement = ast.newForStatement();
			VariableDeclarationFragment vdf = ast.newVariableDeclarationFragment();
			vdf.setName(ast.newSimpleName("i"));
			VariableDeclarationExpression vde = ast.newVariableDeclarationExpression(vdf);
			vde.setType(ast.newSimpleType(ast.newSimpleName(targetType.getName().toString())));
			forStatement.initializers().add(vde);
			
			return forStatement;
		}
	}
	
	private static class ForeachLoopVariableTypeBuilder extends StatementReferenceBuilder {
		@Override
		protected Statement getStatement(TypeDeclaration targetType) {
			/* 
			 * for (A element : new A[]{}) {}
			 */
			
			// build variable
			SingleVariableDeclaration param = ast.newSingleVariableDeclaration();
			param.setType(ast.newArrayType(ast.newSimpleType(ast.newSimpleName(targetType.getName().toString()))));
			param.setName(ast.newSimpleName("element"));
			
			// build expression
			ArrayCreation arrInit = ast.newArrayCreation();
			arrInit.setType(ast.newArrayType(ast.newSimpleType(ast.newSimpleName(targetType.getName().toString()))));
			arrInit.setInitializer(ast.newArrayInitializer());
			
			// build forStatement					
			EnhancedForStatement forStmt = ast.newEnhancedForStatement();
			forStmt.setParameter(param);
			forStmt.setExpression(arrInit);
			forStmt.setBody(ast.newBlock());
			
			return forStmt;
		}
	}
	
	private static class StaticMethodCallBuilder extends StatementReferenceBuilder {
		@Override
		protected Statement getStatement(TypeDeclaration targetType) {
			// must add a dummy static method to the target type
			MethodDeclaration methDecl = BuilderUtil.getDummyMethod(targetType);
			methDecl.setBody(ast.newBlock());
			methDecl.modifiers().add(ast.newModifier(Modifier.ModifierKeyword.STATIC_KEYWORD));
			targetType.bodyDeclarations().add(methDecl);
			
			MethodInvocation methInv = ast.newMethodInvocation();
			methInv.setExpression(ast.newSimpleName(targetType.getName().toString()));
			methInv.setName(ast.newSimpleName(methDecl.getName().toString()));
			
			return ast.newExpressionStatement(methInv);
		}
	}
	
	
	
	/// Expression Reference Builders
	
	private static abstract class ExpressionReferenceBuilder implements RelationshipBuilder {
		public void build(TypeDeclaration a, TypeDeclaration b) {
			
			// find the target method in b
			MethodDeclaration dummyMethod = BuilderUtil.getExistingDummyMethod(b);
			
			// create dummy assignment to hold expression
			VariableDeclarationExpression varDeclExpr = BuilderUtil.getDummyVariable(dummyMethod);			
			Assignment assign = ast.newAssignment();
			assign.setLeftHandSide(varDeclExpr);
			assign.setRightHandSide(getExpression(a));
			
			// add the expression to the method
			dummyMethod.getBody().statements().add(ast.newExpressionStatement(assign));
		}
		protected abstract Expression getExpression(TypeDeclaration targetType);
	}
	
	private static class ConstructorCallBuilder extends ExpressionReferenceBuilder {
		@Override
		protected Expression getExpression(TypeDeclaration targetType) {
			// build instance creation
			ClassInstanceCreation clasInstCrea = ast.newClassInstanceCreation();
			clasInstCrea.setType(ast.newSimpleType((Name)ASTNode.copySubtree(ast, targetType.getName())));
			// find type constructor 
			for (Object declaration : targetType.bodyDeclarations()) {
				if (declaration instanceof MethodDeclaration && ((MethodDeclaration)declaration).isConstructor()) {
					// and add a null argument for each parameter
					for (Object param : ((MethodDeclaration)declaration).parameters()) {
						clasInstCrea.arguments().add(ast.newNullLiteral());
					}
				}
			}	
			return clasInstCrea;
		}
	}

//	private static class AnonymousClassBuilder implements ExpressionReferenceBuilder {
//		public Expression build(TypeDeclaration targetType) {
//			// build anonymous type
//			AnonymousClassDeclaration anonClasDecl = ast.newAnonymousClassDeclaration();
//			
//			// build instance creation
//			ClassInstanceCreation clasInstCrea = (ClassInstanceCreation)(new ConstructorCallBuilder().build(targetType));
//			clasInstCrea.setAnonymousClassDeclaration(anonClasDecl);
//			
//			return clasInstCrea;
//		}
//	}
	
	private static class ArrayCreationBuilder extends ExpressionReferenceBuilder {
		@Override
		protected Expression getExpression(TypeDeclaration targetType) {
			ArrayCreation arrCrea = ast.newArrayCreation();
			arrCrea.setType(ast.newArrayType(ast.newSimpleType(ast.newSimpleName(targetType.getName().toString()))));
			arrCrea.setInitializer(ast.newArrayInitializer());
			
			return arrCrea;
		}
	}
	
	private static class CastBuilder extends ExpressionReferenceBuilder {
		@Override
		protected Expression getExpression(TypeDeclaration targetType) {
			CastExpression castExpression = ast.newCastExpression();
			castExpression.setType(ast.newSimpleType(ast.newName(targetType.getName().toString())));
			castExpression.setExpression(ast.newNullLiteral());
			return castExpression;
		}
	}
	
	private static class InstanceofBuilder extends ExpressionReferenceBuilder {
		@Override
		protected Expression getExpression(TypeDeclaration targetType) {
			InstanceofExpression instanceOfExpression = ast.newInstanceofExpression();
			instanceOfExpression.setLeftOperand(ast.newNullLiteral());
			instanceOfExpression.setRightOperand(ast.newSimpleType(ast.newName(targetType.getName().toString())));
			return instanceOfExpression;
		}
	}
	
	private static class TypeLiteralBuilder extends ExpressionReferenceBuilder {
		@Override
		protected Expression getExpression(TypeDeclaration targetType) {
			TypeLiteral typeLiteral = ast.newTypeLiteral();
			SimpleType typeA = ast.newSimpleType(ast.newName(targetType.getName().toString()));			
			typeLiteral.setType(typeA);	
			return typeLiteral;				
		}
	}
	
	private static class StaticFieldReferenceBuilder extends ExpressionReferenceBuilder {
		@Override
		protected Expression getExpression(TypeDeclaration targetType) {
			// must add a dummy static field to the target type
			FieldDeclaration fieldDecl = BuilderUtil.getDummyField(targetType);
			fieldDecl.setType(ast.newSimpleType(ast.newSimpleName("Object")));
			fieldDecl.modifiers().add(ast.newModifier(Modifier.ModifierKeyword.STATIC_KEYWORD));
			targetType.bodyDeclarations().add(fieldDecl);
			
			FieldAccess fieldAcc = ast.newFieldAccess();
			fieldAcc.setExpression(ast.newSimpleName(targetType.getName().toString()));
			fieldAcc.setName(ast.newSimpleName(((VariableDeclarationFragment)fieldDecl.fragments().get(0)).getName().toString()));
			
			return fieldAcc;
		}
	}
	
	private static class QualifiedThisBuilder extends ExpressionReferenceBuilder {
		@Override
		protected Expression getExpression(TypeDeclaration targetType) {			
			ThisExpression thisExpr = ast.newThisExpression();
			thisExpr.setQualifier(ast.newSimpleName(targetType.getName().toString()));
			return thisExpr;
		}
	}
	
	private RelationshipBuilder builder;
	private RelationshipDirection direction;
	private Reversibility reversibility;
	
	private ReferenceType(
			RelationshipBuilder builder,
			RelationshipDirection direction,
			Reversibility reversibility) {
		this.builder = builder;
		this.direction = direction;
		this.reversibility = reversibility;
	}	
	public RelationshipBuilder getBuilder() {
		return builder;	
	}
	public String getDescription() {		
		return this.toString();
	}
	public Reversibility getReversibility() {
		return this.reversibility;
	}
	public RelationshipDirection getDirection() {
		return this.direction;
	}
	
	/**
	 * @return the list of reference types that can appear in a class-level declaration
	 */
	public ReferenceType[] getDeclarationReferenceTypes() {
		return new ReferenceType[] {
			CONSTRUCTOR_PARAMETER_TYPE,
			FIELD_TYPE,
			PARAMETER_TYPE,
			RETURN_TYPE,
			THROWS_CLAUSE_TYPE
		};
	}
	
	/**
	 * @return the list of reference types that can appear in a method-level statement
	 */
	public ReferenceType[] getStatementReferenceTypes() {
		return new ReferenceType[] {
			CATCH_CLAUSE_PARAMETER_TYPE,
			LOCAL_VARIABLE_TYPE,
			FOR_LOOP_VARIABLE_TYPE,
			FOREACH_LOOP_VARIABLE_TYPE,
			STATIC_METHOD_CALL
		};
	}
	
	/**
	 * @return the list of reference types that can appear in an expression
	 */
	public ReferenceType[] getExpressionReferenceTypes() {
		return new ReferenceType[] {
			ARRAY_CREATION_EXPRESSION,
			CAST_EXPRESSION,
			CONSTRUCTOR_CALL_EXPRESSION,
			INSTANCEOF_EXPRESSION,
			QUALIFIED_THIS_EXPRESSION,
			STATIC_FIELD_REFERENCE_EXPRESSION,
			TYPE_LITERAL_EXPRESSION
		};
	}
	
}