package edu.uiuc.astgen.classrelationships;

import org.testorrery.Chain;
import org.testorrery.CompositeGenerator;
import org.testorrery.IGenerator;
import org.testorrery.Literal;
import org.testorrery.link.IIterationLink;
import org.testorrery.link.LinkCreator;

/**
 * Generator that returns a {@link RelationshipSpecification} that is some combinatoric composition
 *  of the provided {@link LocationType}, {@link InheritanceType}, and {@link ReferenceType} builders.
 */
public class RelationshipSpecificationGenerator extends CompositeGenerator<RelationshipSpecification> {

	/// user-set values
	
	private LocationType[] locationTypes;
	private InheritanceType[] inheritanceTypes;
	private ReferenceType[] referenceTypes;
	
	private boolean enableReverseLocation = false;
	private boolean enableReverseInheritance = false;
	private boolean enableReverseReference = false;
	
	/// generators built on initialization from user-set values
	
	private IGenerator<RelationshipSpecification> locationGen;
	private IGenerator<RelationshipSpecification> inheritanceGen;
	private IGenerator<RelationshipSpecification> referenceGen;
		
	public RelationshipSpecificationGenerator() {
		
	}	
	
	@Override
	protected IIterationLink getInitializedIterator() {				
		locationGen = buildGenerator(locationTypes, LocationType.SEPARATE, enableReverseLocation);
		inheritanceGen = buildGenerator(inheritanceTypes, InheritanceType.NO_INHERITANCE, enableReverseInheritance);
		referenceGen = buildGenerator(referenceTypes, ReferenceType.NO_REFERENCE, enableReverseReference);
		
		return LinkCreator.createCombinatoricLink(
				locationGen, 
				inheritanceGen,
				referenceGen);
	}
	
	/**
	 * Builds a generator from the given list of {@link RelationshipSpecification}s. 
	 * @param specs The list of {@link RelationshipSpecification}s from which to build a generator. Can be null and empty.
	 * @param defaultSpec The default specification if the given list is null or empty 
	 * @param enableReverse Whether or not to include reversed specifications in the generator
	 */
	private static <SpecType extends RelationshipSpecification> IGenerator<RelationshipSpecification> buildGenerator(SpecType[] specs, SpecType defaultSpec, boolean enableReverse) {		
		if (specs == null || specs.length == 0) {
			// default value if the user did not set anything
			return new Literal<RelationshipSpecification>(defaultSpec);
		}
		// build sequence from provided specifications
		Chain<RelationshipSpecification> specSeq = new Chain<RelationshipSpecification>();
		for (SpecType spec : specs) {
			specSeq.add(spec);
			if (enableReverse && spec.getReversibility() == Reversibility.IS_REVERSIBLE) {
				specSeq.add(new ReversedRelationship(spec));
			}
		}
		return specSeq;
	}

	public void setEnableReverseLocation(boolean enableReverse) {
		enableReverseLocation = enableReverse;
	}

	public void setEnableReverseInheritance(boolean enableReverse) {
		enableReverseInheritance = enableReverse;
	}

	public void setEnableReverseReference(boolean enableReverse) {
		enableReverseReference = enableReverse;
	}
	
	public void setEnableReverse(boolean enableReverse) {
		setEnableReverseInheritance(enableReverse);
		setEnableReverseLocation(enableReverse);
		setEnableReverseReference(enableReverse);
	}
		
	public void setInheritance(InheritanceType...inheritanceTypes) {
		this.inheritanceTypes = inheritanceTypes;
	}
	
	public void setInheritanceAll() {
		setInheritance(InheritanceType.values());
	}

	public void setLocation(LocationType... locationTypes) {
		this.locationTypes = locationTypes;
	}
	
	public void setLocationAll() {
		setLocation(LocationType.values());
	}
	
	public void setReference(ReferenceType... referenceTypes) {
		this.referenceTypes = referenceTypes;
	}
	
	public void setReferenceAll() {
		setReference(ReferenceType.values());
	}

	@Override
	protected RelationshipSpecification generateCurrent() {
		RelationshipSpecification locationSpec = locationGen.current();
		RelationshipSpecification inheritanceSpec = inheritanceGen.current();
		RelationshipSpecification referenceSpec = referenceGen.current();
		
		return new CompositeRelationship(locationSpec, inheritanceSpec, referenceSpec);		
	}

}
