package edu.uiuc.astgen.fieldreferences;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;
import org.eclipse.jdt.core.dom.Modifier.ModifierKeyword;
import org.testorrery.Chain;
import org.testorrery.CompositeGenerator;
import org.testorrery.IGenerator;
import org.testorrery.link.IIterationLink;
import org.testorrery.link.LinkCreator;

import edu.uiuc.astgen.ASTNodeGeneratorBase;
import edu.uiuc.astgen.classrelationships.BuilderUtil;
import edu.uiuc.astgen.classrelationships.RelationshipSpecification;
import edu.uiuc.astgen.classrelationships.RelationshipSpecificationGenerator;
import edu.uiuc.astgen.nestedexpressions.NestedExpressionBuilder;
import edu.uiuc.astgen.nestedexpressions.TypedFieldExpressionSpec;
import edu.uiuc.astgen.nestedexpressions.TypedFieldExpressionSpecGenerator;
import edu.uiuc.astgen.statements.StatementWithExpressionBuilder;
import edu.uiuc.astgen.statements.StatementWithExpressionBuilderGenerator;

/**
 * Generates a compilation unit with two classes, one of which declares a field that the other references in some way.
 */
public class DualClassFieldReferenceGenerator extends CompositeGenerator<CompilationUnit> {
	
	/**
	 * The target field
	 */
	private IGenerator<FieldDeclaration> fieldDeclGen;
	/**
	 * The way the field is being referenced
	 */
	private IGenerator<FieldReferenceType> fieldRefTypeGen;
	/**
	 * The generator for relating two classes
	 */
	private RelationshipSpecificationGenerator relationshipSpecificationGen; 	
	/**
	 * The nested expression builder with other metadata
	 */
	private IGenerator<TypedFieldExpressionSpec> expressionSpecGen;
	/**
	 * The builder for the statement in which the expression will appear
	 */
	private IGenerator<StatementWithExpressionBuilder> stmtGen; 
		
	private static final AST ast = ASTNodeGeneratorBase.ast();
	
	public DualClassFieldReferenceGenerator(IGenerator<FieldDeclaration> fieldDeclGen) {
		this.fieldDeclGen = fieldDeclGen;
	}
	
	public void setRelationshipSpecificationGen(
			RelationshipSpecificationGenerator relationshipSpecificationGen) {
		this.relationshipSpecificationGen = relationshipSpecificationGen;
	}
	
	@Override
	protected IIterationLink getInitializedIterator() {
		if (relationshipSpecificationGen == null) {
			relationshipSpecificationGen = new RelationshipSpecificationGenerator();
			relationshipSpecificationGen.setLocationAll();
			relationshipSpecificationGen.setInheritanceAll();	
		}
		
		fieldRefTypeGen = new Chain<FieldReferenceType>(FieldReferenceType.values());
				
		stmtGen = StatementWithExpressionBuilderGenerator.allExpressionTypes();
		
		expressionSpecGen = TypedFieldExpressionSpecGenerator.newGenerator(fieldDeclGen);
		
		IIterationLink link = LinkCreator.createCombinatoricLink(
				//fieldDeclGen, iterated as part of expressionSpecGen 
				relationshipSpecificationGen, 
				fieldRefTypeGen, 
				expressionSpecGen,
				stmtGen);
				
		return link;
	}
	
	@Override
	protected CompilationUnit generateCurrent() {
		TypedFieldExpressionSpec expressionSpec = expressionSpecGen.current();
		FieldDeclaration fieldDecl = (FieldDeclaration)ASTNode.copySubtree(ast, expressionSpec.getFieldDecl());
		RelationshipSpecification relSpec = relationshipSpecificationGen.current();
		FieldReferenceType fieldRefType = fieldRefTypeGen.current();
		NestedExpressionBuilder nestedExprBuilder = expressionSpec.getExpressionBuilder();
		StatementWithExpressionBuilder stmtBuilder = stmtGen.current();
		
		SimpleName fieldName = ast.newSimpleName(((VariableDeclarationFragment)fieldDecl.fragments().get(0)).getName().toString());
		
		// build first class
		TypeDeclaration classA = ast.newTypeDeclaration();
		classA.setName(ast.newSimpleName("A"));
		classA.modifiers().add(ast.newModifier(ModifierKeyword.PUBLIC_KEYWORD)); // public since class A is the main type in the compilation unit
		
		// build second class
		TypeDeclaration classB = ast.newTypeDeclaration();
		classB.setName(ast.newSimpleName("B"));
		
		// add field declaration to A
		classA.bodyDeclarations().add(fieldDecl);
		
		// build field reference (i.e. field, this.field, super.field, etc.)
		Expression fieldRef = fieldRefType.getBuilder().build(fieldName, classA);
		
		// build dummy variable for expressions that need one
		String dummyVariableName = "dummyVariable";
		VariableDeclarationFragment varDeclFrag = ast.newVariableDeclarationFragment();		
		varDeclFrag.setName(ast.newSimpleName(dummyVariableName));
		varDeclFrag.setInitializer((Expression)ASTNode.copySubtree(ast, expressionSpec.getInitializingExpression()));
		VariableDeclarationStatement varDeclStmt = ast.newVariableDeclarationStatement(varDeclFrag);
		varDeclStmt.setType((Type)ASTNode.copySubtree(ast, fieldDecl.getType()));
		
		// build referencing expression (containing field reference)
		Expression referencingExpression = nestedExprBuilder.build(fieldRef, ast.newSimpleName(dummyVariableName));
		
		// build referencing statement
		Statement stmt = stmtBuilder.build(referencingExpression);
		
		// build referencing method	
		Block methBody = ast.newBlock();
		methBody.statements().add(varDeclStmt);
		methBody.statements().add(stmt);		
		MethodDeclaration methDecl = BuilderUtil.getDummyMethod(classB);
		methDecl.setBody(methBody);
		
		// add referencing method to B
		classB.bodyDeclarations().add(methDecl);
		
		// link the classes together
		relSpec.getBuilder().build(classA, classB);		
		
		CompilationUnit cu = ast.newCompilationUnit();
		cu.types().add(classA);
		if (classB.getParent() == null) {
			cu.types().add(classB);
		}
		
		return cu;
	}
}
