package edu.uiuc.astgen.fieldreferences;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ClassInstanceCreation;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.FieldAccess;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.SuperFieldAccess;
import org.eclipse.jdt.core.dom.ThisExpression;
import org.eclipse.jdt.core.dom.TypeDeclaration;

import edu.uiuc.astgen.ASTNodeGeneratorBase;

public enum FieldReferenceType {
	/**
	 * fieldName
	 */
	UNQUALIFIED(new UnqualifiedBuilder()),
	/**
	 * new A().fieldName
	 */
	NEW_INSTANCE(new NewInstanceBuilder()),
	/**
	 * this.fieldName
	 */
	THIS(new ThisBuilder()),
	/**
	 * super.fieldName
	 */
	SUPER(new SuperBuilder()),
	/**
	 * A.this.fieldName
	 */
	QUALIFIED_THIS(new QualifiedThisBuilder()),
	/** 
	 * No reference type
	 */
	NO_REFERENCE_TYPE(new NoReferenceType());
	
	public interface FieldReferenceBuilder {
		Expression build(SimpleName fieldName, TypeDeclaration container);
	}
	
	private static final AST ast = ASTNodeGeneratorBase.ast();
	
	private static class NoReferenceType implements FieldReferenceBuilder {
		public Expression build(SimpleName fieldName, TypeDeclaration container) {
			return ast.newNullLiteral();
		}		
	}
	
	private static class UnqualifiedBuilder implements FieldReferenceBuilder {
		public Expression build(SimpleName fieldName, TypeDeclaration container) {
			if (fieldName.getParent() == null) {
				return fieldName;
			}			
			return ast.newSimpleName(fieldName.toString());
		}
	}
	
	private static class NewInstanceBuilder implements FieldReferenceBuilder {
		public Expression build(SimpleName fieldName, TypeDeclaration container) {
			ClassInstanceCreation instExpr = ast.newClassInstanceCreation();
			instExpr.setType(ast.newSimpleType(ast.newSimpleName(container.getName().toString())));
			// find constructor method
			MethodDeclaration methDecl;
			for (Object bodyDecl : container.bodyDeclarations()) {
				if (bodyDecl instanceof MethodDeclaration) {
					methDecl = (MethodDeclaration)bodyDecl;
					if (methDecl.isConstructor()) {
						// add a null argument for each parameter
						for (Object param : methDecl.parameters()) {
							instExpr.arguments().add(ast.newNullLiteral());
						}
					}
				}
			}
			
			FieldAccess fieldAcc = ast.newFieldAccess();
			fieldAcc.setExpression(instExpr);
			fieldAcc.setName(ast.newSimpleName(fieldName.toString()));
			
			return fieldAcc;
		}
	}
	
//	private static class VariableBuilder implements FieldReferenceBuilder {
//		public Expression build(SimpleName fieldName, TypeDeclaration container) {
//			FieldAccess fieldAcc = ast.newFieldAccess();
//			fieldAcc.setExpression(ast.newSimpleName("DummyInstanceOf" + container.getName().toString()));
//			fieldAcc.setName(ast.newSimpleName(fieldName.toString()));
//			
//			return fieldAcc;
//		}
//	}
	
	private static class ThisBuilder implements FieldReferenceBuilder {
		public Expression build(SimpleName fieldName, TypeDeclaration container) {
			FieldAccess fieldAcc = ast.newFieldAccess();
			fieldAcc.setExpression(ast.newThisExpression());
			fieldAcc.setName(ast.newSimpleName(fieldName.toString()));
			
			return fieldAcc;
		}
	}
	
	private static class SuperBuilder implements FieldReferenceBuilder {
		public Expression build(SimpleName fieldName, TypeDeclaration container) {
			SuperFieldAccess superFieldAcc = ast.newSuperFieldAccess();
			superFieldAcc.setName(ast.newSimpleName(fieldName.toString()));
			
			return superFieldAcc;
		}
	}
	
	private static class QualifiedThisBuilder implements FieldReferenceBuilder {
		public Expression build(SimpleName fieldName, TypeDeclaration container) {
			ThisExpression thisExpr = ast.newThisExpression();
			thisExpr.setQualifier(ast.newSimpleName(container.getName().toString()));
			
			FieldAccess fieldAcc = ast.newFieldAccess();
			fieldAcc.setExpression(thisExpr);
			fieldAcc.setName(ast.newSimpleName(fieldName.toString()));
			
			return fieldAcc;
		}
	}	
	
	private FieldReferenceBuilder builder;
	
	private FieldReferenceType(FieldReferenceBuilder builder) {
		this.builder = builder;
	}
	public FieldReferenceBuilder getBuilder() {
		return builder;
	}
}
