package edu.uiuc.astgen.fieldreferences;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;
import org.eclipse.jdt.core.dom.Modifier.ModifierKeyword;
import org.testorrery.Chain;
import org.testorrery.CompositeGenerator;
import org.testorrery.IGenerator;
import org.testorrery.link.IIterationLink;
import org.testorrery.link.LinkCreator;

import edu.uiuc.astgen.ASTNodeGeneratorBase;
import edu.uiuc.astgen.classrelationships.BuilderUtil;
import edu.uiuc.astgen.methodrelationships.SingleClassMethodReferenceGenerator;
import edu.uiuc.astgen.nestedexpressions.NestedExpressionBuilder;
import edu.uiuc.astgen.nestedexpressions.NotNestedBuilder;
import edu.uiuc.astgen.nestedexpressions.ParenthesesBuilder;
import edu.uiuc.astgen.nestedexpressions.TypedFieldExpressionSpec;
import edu.uiuc.astgen.nestedexpressions.TypedFieldExpressionSpecGenerator;
import edu.uiuc.astgen.statements.StatementWithExpressionBuilder;
import edu.uiuc.astgen.statements.StatementWithExpressionBuilderGenerator;

public class SingleClassFieldReferenceGenerator extends	CompositeGenerator<TypeDeclaration> {

	/**
	 * The class containing the target field
	 */
	private String className;
	/**
	 * The target field.
	 */
	private IGenerator<FieldDeclaration> fieldDeclGen; 
	/**
	 * The way the field is being referenced (i.e. this.field, super.field, etc)
	 */
	private IGenerator<FieldReferenceType> fieldRefTypeGen; 
	/**
	 *  The way the field is enclosed in a nested expression
	 */
	private IGenerator<NestedExpressionBuilder> fieldRefExpr;
	/**
	 * The nested expression builder with other metadata
	 * 
	 * This TypedFieldExpressionSpec stuff is a nasty hack to get expressions to be type-correct.
	 * See {@link SingleClassMethodReferenceGenerator#iterationIsValid for a better way to do it.
	 */
	private IGenerator<TypedFieldExpressionSpec> expressionSpecGen;
	/**
	 * The builder for the statement in which the expression will appear
	 */
	private IGenerator<StatementWithExpressionBuilder> stmtGen; 
		
	private static final AST ast = ASTNodeGeneratorBase.ast();
	
	public SingleClassFieldReferenceGenerator(String className, IGenerator<FieldDeclaration> fieldDeclGen) {
		this.className = className;
		this.fieldDeclGen = fieldDeclGen;
	}
	
	@Override
	protected IIterationLink getInitializedIterator() {
		fieldRefTypeGen = new Chain<FieldReferenceType>(
				FieldReferenceType.NEW_INSTANCE, 
				FieldReferenceType.QUALIFIED_THIS, 
				FieldReferenceType.THIS,
				FieldReferenceType.UNQUALIFIED);
		
		expressionSpecGen = TypedFieldExpressionSpecGenerator.newGenerator(fieldDeclGen);
		
		stmtGen = StatementWithExpressionBuilderGenerator.allExpressionTypes();
		fieldRefExpr = new Chain<NestedExpressionBuilder>(new NotNestedBuilder(), new ParenthesesBuilder());		
		
		IIterationLink link = LinkCreator.createCombinatoricLink(
				//fieldDeclGen, part of expressionSpecGen 
				fieldRefTypeGen, 
				expressionSpecGen,
				fieldRefExpr,
				stmtGen);
				
		return link;
	}
	
	@Override
	protected TypeDeclaration generateCurrent() {
		TypedFieldExpressionSpec expressionSpec = expressionSpecGen.current();
		FieldDeclaration fieldDecl = (FieldDeclaration)ASTNode.copySubtree(ast, expressionSpec.getFieldDecl());
		NestedExpressionBuilder nestedExprBuilder = expressionSpec.getExpressionBuilder();
		Expression dummyInitializer = (Expression)ASTNode.copySubtree(ast, expressionSpec.getInitializingExpression());
		
		FieldReferenceType fieldRefType = fieldRefTypeGen.current();
		NestedExpressionBuilder fieldRefExprBuilder = fieldRefExpr.current();
		StatementWithExpressionBuilder stmtBuilder = stmtGen.current();
		
		SimpleName fieldName = ast.newSimpleName(((VariableDeclarationFragment)fieldDecl.fragments().get(0)).getName().toString());
		
		// build class
		TypeDeclaration classA = ast.newTypeDeclaration();
		classA.setName(ast.newSimpleName(className));
		classA.modifiers().add(ast.newModifier(ModifierKeyword.PUBLIC_KEYWORD)); // public since class A is the main type in the compilation unit
	
		// add field declaration to class
		classA.bodyDeclarations().add(fieldDecl);
		
		// build field reference (i.e. field, this.field, super.field, etc.)
		Expression fieldRef = fieldRefType.getBuilder().build(fieldName, classA);
		
		// build dummy variable for expressions that need one
		String dummyVariableName = "dummyVariable";
		VariableDeclarationFragment varDeclFrag = ast.newVariableDeclarationFragment();		
		varDeclFrag.setName(ast.newSimpleName(dummyVariableName));
		varDeclFrag.setInitializer(dummyInitializer);
		VariableDeclarationStatement varDeclStmt = ast.newVariableDeclarationStatement(varDeclFrag);
		varDeclStmt.setType((Type)ASTNode.copySubtree(ast, fieldDecl.getType()));
		
		// build referencing expression (containing field reference)
		Expression fieldExpr = fieldRefExprBuilder.build(fieldRef, null);
				
		// build expression containing the field expression
		Expression referencingExpression = nestedExprBuilder.build(fieldExpr, ast.newSimpleName(dummyVariableName));
		
		// build referencing statement
		Statement stmt = stmtBuilder.build(referencingExpression);
		
		// build referencing method	
		Block methBody = ast.newBlock();
		methBody.statements().add(varDeclStmt);
		methBody.statements().add(stmt);		
		MethodDeclaration methDecl = BuilderUtil.getDummyMethod(classA);
		methDecl.setBody(methBody);
		
		classA.bodyDeclarations().add(methDecl);
				
		return classA;
	}

}
