package edu.uiuc.astgen.methodrelationships;

import java.util.List;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Assignment;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.ExpressionStatement;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.ReturnStatement;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.StringLiteral;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;
import org.testorrery.Chain;
import org.testorrery.FilteredCompositeGenerator;
import org.testorrery.IGenerator;
import org.testorrery.link.IIterationLink;
import org.testorrery.link.LinkCreator;

import edu.uiuc.astgen.ASTNodeGeneratorBase;
import edu.uiuc.astgen.classrelationships.InheritanceType;
import edu.uiuc.astgen.classrelationships.RelationshipSpecification;
import edu.uiuc.astgen.classrelationships.RelationshipSpecificationGenerator;
import edu.uiuc.astgen.fieldreferences.FieldReferenceType;
import edu.uiuc.astgen.nestedexpressions.NestedExpressionBuilder;
import edu.uiuc.astgen.nestedexpressions.NestedExpressionBuilderGenerator;
import edu.uiuc.astgen.nestedexpressions.NotNestedBuilder;
import edu.uiuc.astgen.nestedexpressions.OperandType;
import edu.uiuc.astgen.nestedexpressions.ParenthesesBuilder;
import edu.uiuc.astgen.statements.ExpressionStatementBuilder;
import edu.uiuc.astgen.statements.StatementWithExpressionBuilder;
import edu.uiuc.astgen.statements.StatementWithExpressionBuilderGenerator;

public class DualClassMethodParentGenerator extends FilteredCompositeGenerator<CompilationUnit> {

	private IGenerator<MethodDeclaration> methDeclGen;
	private MethodRelationshipBuilderGenerator methRelGen;
	private MethodReferenceBuilderGenerator methRefGen;
	private IGenerator<NestedExpressionBuilder> methExprGen;
	private IGenerator<NestedExpressionBuilder> nestedExprGen;
	private IGenerator<StatementWithExpressionBuilder> exprInStmtGen;
	/**
	 * The generator for relating two classes
	 */
	private RelationshipSpecificationGenerator relationshipSpecificationGen;
	private IGenerator<FieldDeclaration> fieldDeclGen;
	/**
	 * The way the field is being referenced (i.e. this.field, super.field, etc)
	 */
	private IGenerator<FieldReferenceType> fieldRefTypeGen; 	
	private String classParent;
	private String classChild;
	
	/** the generator for locating 2 methods in class A and B */
	Chain<String> typeLocationMethodsChain;
	
	
	private static AST ast = ASTNodeGeneratorBase.ast();
	
	public DualClassMethodParentGenerator(IGenerator<FieldDeclaration> fieldDeclGen, IGenerator<MethodDeclaration> methDeclGen, String classParent, String classChild) {
		this.fieldDeclGen = fieldDeclGen;
		this.methDeclGen = methDeclGen;
		this.classParent = classParent;
		this.classChild = classChild;
	}
	
	@Override
	protected IIterationLink getInitializedIterator() {
		typeLocationMethodsChain =  new Chain<String>(classParent, classChild);
		fieldRefTypeGen = new Chain<FieldReferenceType>(
				FieldReferenceType.NO_REFERENCE_TYPE,
				FieldReferenceType.NEW_INSTANCE, 
				FieldReferenceType.QUALIFIED_THIS, 
				FieldReferenceType.THIS,
				FieldReferenceType.UNQUALIFIED);
		
		if (methRelGen == null) {
			methRelGen = new MethodRelationshipBuilderGenerator();
		}
		if (methRefGen == null) {
			methRefGen = new MethodReferenceBuilderGenerator();
		}
		if (methExprGen == null) {			
			methExprGen = NestedExpressionBuilderGenerator.parenNoParenAllOperandTypes();
		}
		if (nestedExprGen == null) {
			nestedExprGen = NestedExpressionBuilderGenerator.allOperandTypes();
		}
		if (exprInStmtGen == null) {
			exprInStmtGen = StatementWithExpressionBuilderGenerator.allExpressionTypes();
		}
		if (relationshipSpecificationGen == null) {
			relationshipSpecificationGen = new RelationshipSpecificationGenerator();
			relationshipSpecificationGen.setLocationAll();
			InheritanceType[] inheritanceTypes = new InheritanceType[1];
			inheritanceTypes[0] = InheritanceType.SUBCLASS;
			//relationshipSpecificationGen.setInheritanceAll();
			relationshipSpecificationGen.setInheritance(inheritanceTypes);	
		}		
		return LinkCreator.createCombinatoricLink(
				typeLocationMethodsChain,
				relationshipSpecificationGen,
				methDeclGen, 
				methRelGen, 
				methRefGen,
				methExprGen,
				nestedExprGen,
				exprInStmtGen,
				//typeBodyMethodDeclarationChain, 
				fieldRefTypeGen,
				fieldDeclGen);
	}
	
	@Override
	protected boolean iterationIsValid() {
		MethodDeclaration methDecl = (MethodDeclaration)ASTNode.copySubtree(ast, methDeclGen.current());
		MethodRelationshipBuilder methRelBuilder = methRelGen.current();
		MethodReferenceBuilder methRefBuilder = methRefGen.current();
		NestedExpressionBuilder methExprBuilder = methExprGen.current();
		NestedExpressionBuilder nestedExprBuilder = nestedExprGen.current();
		StatementWithExpressionBuilder exprInStmtBuilder = exprInStmtGen.current();		
		
		if (methDecl.parameters().size() == 0
			&& methRelBuilder.willRemoveParam()) {
			// disallow overloading with fewer parameters
			return false;
		}
		if (methRelBuilder.requiresSuper()) {
			// disallow overriding
			return false;
		}		
		if (methRefBuilder.isSuper()) {
			// disallow super.m() calls since this will only have one class
			return false;
		}
		if (!methExprBuilder.applicableTo(OperandType.fromDomType(methDecl.getReturnType2()))) {
			// disallow expressions that are not applicable to the method's return type
			return false;
		}
		if (nestedExprBuilder.isAssignment()) {
			// cannot assign to a method
			return false;
		}
		if (!nestedExprBuilder.applicableTo(methExprBuilder.outputType())) {
			// if output type of nested expression does not match the input type of the outer expression
			return false;
		}
		if (!exprInStmtBuilder.applicableTo(nestedExprBuilder.outputType())) {
			// if output type of outer expression does not match the input type of the statement
			return false;
		}
		if (methExprBuilder instanceof ParenthesesBuilder 
				&& exprInStmtBuilder instanceof ExpressionStatementBuilder) {
			// disallow parenthesized expression statement
			// (expr);
			return false;
		}
		if (methExprBuilder instanceof NotNestedBuilder 
				&& nestedExprBuilder instanceof ParenthesesBuilder
				&& exprInStmtBuilder instanceof ExpressionStatementBuilder) {
			// disallow parenthesized expression statement
			// (expr);
			return false;
		}
		return true;
	}

	@Override
	protected CompilationUnit generateCurrent() {
		String locateMethodClass = typeLocationMethodsChain.current();
		RelationshipSpecification relSpec = relationshipSpecificationGen.current();
		MethodDeclaration methDecl = (MethodDeclaration)ASTNode.copySubtree(ast, methDeclGen.current());
		MethodRelationshipBuilder methRelBuilder = methRelGen.current();
		MethodReferenceBuilder methRefBuilder = methRefGen.current();
		NestedExpressionBuilder methExprBuilder = methExprGen.current();
		NestedExpressionBuilder nestedExprBuilder = nestedExprGen.current();
		StatementWithExpressionBuilder exprInStmtBuilder = exprInStmtGen.current();
		FieldDeclaration fieldDecl = fieldDeclGen.current();
		FieldReferenceType fieldRefType = fieldRefTypeGen.current();
	
		TypeDeclaration classA = ast.newTypeDeclaration();
		classA.setName(ast.newSimpleName(classParent));		
		classA.bodyDeclarations().add(fieldDecl);
		classA.bodyDeclarations().add(methDecl);
		classA.modifiers().add(ast.newModifier(Modifier.ModifierKeyword.PUBLIC_KEYWORD));
		
		
		// build second class
		TypeDeclaration classB = ast.newTypeDeclaration();
		classB.setName(ast.newSimpleName(classChild));		
		
		SimpleName fieldName = ast.newSimpleName(((VariableDeclarationFragment)fieldDecl.fragments().get(0)).getName().toString());
		// build field reference (i.e. field, this.field, super.field, etc.)
		if (!fieldRefType.toString().equals("NO_REFERENCE_TYPE")) {
			Expression fieldRef = fieldRefType.getBuilder().build(fieldName, classA);		
			Assignment assignmentExp = ast.newAssignment();
			assignmentExp.setLeftHandSide(fieldRef);
			assignmentExp.setOperator(Assignment.Operator.ASSIGN);
			Expression initializer =null;
			if (fieldDecl.getType().toString().equals("int") || fieldDecl.getType().toString().equals("Integer")){
				initializer = ast.newNumberLiteral("10");			
			} else if (fieldDecl.getType().toString().equals("boolean") || fieldDecl.getType().toString().equals("Boolean")) {
				initializer = ast.newBooleanLiteral(true);
			} else if (fieldDecl.getType().toString().equals("String")) {
				StringLiteral s = ast.newStringLiteral();
				s.setEscapedValue("\"abcd\"");
				initializer = s;
			} else {
				initializer = ast.newNullLiteral();
			}
			assignmentExp.setRightHandSide(initializer);
			ExpressionStatement expStatem = ast.newExpressionStatement(assignmentExp);
			methDecl.getBody().statements().add((ExpressionStatement)ASTNode.copySubtree(ast, expStatem));
		}
		String methodName = methDecl.getName().toString();
		String newMethodName = methodName + "Prime";
		String dummyVarName = "dummyVariable";		
		
		MethodDeclaration newMethDecl = null;
		if (locateMethodClass.equals(classParent))
			newMethDecl = methRelBuilder.build(newMethodName, methDecl, classA);
		else 
			newMethDecl = methRelBuilder.build(newMethodName, methDecl, classB);
		Expression methRefExpr = methRefBuilder.build(methDecl, classA);
		Expression methExpr = methExprBuilder.build(methRefExpr, ast.newSimpleName("unneeded")); 
		Expression nestedExpr = nestedExprBuilder.build(methExpr, ast.newSimpleName(dummyVarName));
		Statement stmt = exprInStmtBuilder.build(nestedExpr);
		List newMethStmts = newMethDecl.getBody().statements();
		
		if (!methDecl.getReturnType2().toString().equals("void")) {
			VariableDeclarationFragment varDeclFrag = ast.newVariableDeclarationFragment();
			varDeclFrag.setName(ast.newSimpleName(dummyVarName));
			varDeclFrag.setInitializer(ast.newNullLiteral());
			VariableDeclarationStatement dummyVarDecl = ast.newVariableDeclarationStatement(varDeclFrag);
			dummyVarDecl.setType((Type)ASTNode.copySubtree(ast, methDecl.getReturnType2()));
			newMethStmts.add(dummyVarDecl);						
		}
		newMethStmts.add(stmt);
		
		if (!methDecl.getReturnType2().toString().equals("void")) {
			ReturnStatement methReturn = ast.newReturnStatement();
			methReturn.setExpression(OperandType.fromDomType(methDecl.getReturnType2()).dummyExpression());
			methDecl.getBody().statements().add(methReturn);
			
			ReturnStatement newMethReturn = ast.newReturnStatement();
			newMethReturn.setExpression(ast.newSimpleName(dummyVarName));
			newMethStmts.add(newMethReturn);
		}
		
		// link the classes together
		relSpec.getBuilder().build(classA, classB);		
		
		CompilationUnit cu = ast.newCompilationUnit();
		cu.types().add(classA);
		if (classB.getParent() == null) {
			cu.types().add(classB);
		}
	
		return cu;
	}
	
}