package edu.uiuc.astgen.methodrelationships;

import java.util.List;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ClassInstanceCreation;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.SuperMethodInvocation;
import org.eclipse.jdt.core.dom.ThisExpression;
import org.eclipse.jdt.core.dom.TypeDeclaration;

import edu.uiuc.astgen.ASTNodeGeneratorBase;

public abstract class MethodReferenceBuilder {
	protected AST ast = ASTNodeGeneratorBase.ast();
	
	public abstract Expression build(MethodDeclaration referencedMethod, TypeDeclaration container);
	
	protected MethodInvocation makeInvocation(MethodDeclaration referencedMethod) {
		MethodInvocation newMethod = ast.newMethodInvocation();
		newMethod.setName(ast.newSimpleName(referencedMethod.getName().toString()));
		makeArguments(referencedMethod, newMethod.arguments());
		return newMethod;
	}

	protected void makeArguments(MethodDeclaration referencedMethod, List args) {
		for (Object parameter : referencedMethod.parameters()) {
			args.add(ast.newNullLiteral());
		}
	}

	public boolean isSuper() {
		return false;
	}
}

class UnqualifiedBuilder extends MethodReferenceBuilder {
	public Expression build(MethodDeclaration referencedMethod, TypeDeclaration container) {	
		return makeInvocation(referencedMethod);
	}
}

class NewInstanceBuilder extends MethodReferenceBuilder {
	public Expression build(MethodDeclaration referencedMethod, TypeDeclaration container) {
		ClassInstanceCreation instExpr = ast.newClassInstanceCreation();
		instExpr.setType(ast.newSimpleType(ast.newSimpleName(container.getName().toString())));
		// find constructor method
		MethodDeclaration methDecl;
		for (Object bodyDecl : container.bodyDeclarations()) {
			if (bodyDecl instanceof MethodDeclaration) {
				methDecl = (MethodDeclaration)bodyDecl;
				if (methDecl.isConstructor()) {
					// add a null argument for each parameter
					for (Object param : methDecl.parameters()) {
						instExpr.arguments().add(ast.newNullLiteral());
					}
				}
			}
		}
		
		MethodInvocation methInv = makeInvocation(referencedMethod);
		methInv.setExpression(instExpr);
		
		return methInv;
	}
}

class ThisBuilder extends MethodReferenceBuilder {
	public Expression build(MethodDeclaration referencedMethod, TypeDeclaration container) {
		MethodInvocation methInv = makeInvocation(referencedMethod);
		methInv.setExpression(ast.newThisExpression());		
		return methInv;
	}
}

class SuperBuilder extends MethodReferenceBuilder {
	public Expression build(MethodDeclaration referencedMethod, TypeDeclaration container) {
		SuperMethodInvocation superMethInv = ast.newSuperMethodInvocation();
		superMethInv.setName(ast.newSimpleName(referencedMethod.getName().toString()));
		makeArguments(referencedMethod, superMethInv.arguments());
		return superMethInv;
	}
	@Override
	public boolean isSuper() {
		return true;
	}
}

class QualifiedThisBuilder extends MethodReferenceBuilder {
	public Expression build(MethodDeclaration referencedMethod, TypeDeclaration container) {
		ThisExpression thisExpr = ast.newThisExpression();
		thisExpr.setQualifier(ast.newSimpleName(container.getName().toString()));
		
		MethodInvocation methInv = makeInvocation(referencedMethod);
		methInv.setExpression(thisExpr);
		
		return methInv;
	}
}	
