package edu.uiuc.astgen.methodrelationships;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.PrimitiveType;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;
import org.testorrery.FilteredCompositeGenerator;
import org.testorrery.IGenerator;
import org.testorrery.link.IIterationLink;
import org.testorrery.link.LinkCreator;

import edu.uiuc.astgen.ASTNodeGeneratorBase;
import edu.uiuc.astgen.nestedexpressions.NestedExpressionBuilder;
import edu.uiuc.astgen.nestedexpressions.NestedExpressionBuilderGenerator;
import edu.uiuc.astgen.nestedexpressions.OperandType;
import edu.uiuc.astgen.statements.StatementWithExpressionBuilder;
import edu.uiuc.astgen.statements.StatementWithExpressionBuilderGenerator;

public class MethodWithParameterReferenceGenerator extends FilteredCompositeGenerator<MethodDeclaration> {

	private String methodName;
	private IGenerator<SingleVariableDeclaration> paramGen;
	private IGenerator<NestedExpressionBuilder> nestedGen1;
	private IGenerator<NestedExpressionBuilder> nestedGen2;
	private IGenerator<StatementWithExpressionBuilder> exprStmtGen;
	private IGenerator<StatementInBlockBuilder> methBodyGen;
	
	public MethodWithParameterReferenceGenerator(String methodName, IGenerator<SingleVariableDeclaration> paramGen) {
		this.methodName = methodName;
		this.paramGen = paramGen;
	}

	@Override
	protected IIterationLink getInitializedIterator() {
		if (nestedGen1 == null) {
			nestedGen1 = NestedExpressionBuilderGenerator.parenNoParenAllOperandTypes();
		}
		if (nestedGen2 == null) {
			nestedGen2 = NestedExpressionBuilderGenerator.allOperandTypes();			
		}
		if (exprStmtGen == null) {
			exprStmtGen = StatementWithExpressionBuilderGenerator.allExpressionTypes();
		}
		if (methBodyGen == null) {
			methBodyGen = new StatementInBlockBuilderGenerator();
		}
		return LinkCreator.createCombinatoricLink(
				paramGen, 				 
				nestedGen2,
				nestedGen1,
				exprStmtGen,
				methBodyGen);
	}

	@Override
	protected boolean iterationIsValid() {
		SingleVariableDeclaration param = paramGen.current();
		NestedExpressionBuilder exprBuilder1 = nestedGen1.current();
		NestedExpressionBuilder exprBuilder2 = nestedGen2.current();
		StatementWithExpressionBuilder stmtBuilder = exprStmtGen.current();
		StatementInBlockBuilder methBodyBuilder = methBodyGen.current();
		
		if (!exprBuilder1.applicableTo(OperandType.fromDomType(param.getType()))) {
			return false;			
		}
		if (!exprBuilder2.applicableTo(exprBuilder1.outputType())) {
			return false;
		}
		if (!stmtBuilder.applicableTo(exprBuilder2.outputType())) {
			return false;
		}
		if (methBodyBuilder.requiresFinal() && (exprBuilder1.isAssignment() || exprBuilder2.isAssignment())) {
			return false;
		}
		
		return true;
	}
	
	@Override
	protected MethodDeclaration generateCurrent() {		
		SingleVariableDeclaration param = paramGen.current();
		NestedExpressionBuilder exprBuilder1 = nestedGen1.current();
		NestedExpressionBuilder exprBuilder2 = nestedGen2.current();
		StatementWithExpressionBuilder stmtBuilder = exprStmtGen.current();
		StatementInBlockBuilder methBodyBuilder = methBodyGen.current();
		
		AST ast = ASTNodeGeneratorBase.ast();
		
		// make dummy variable
		final String dummyName = "dummyVariable";
		VariableDeclarationFragment varDeclFrag = ast.newVariableDeclarationFragment();
		varDeclFrag.setName(ast.newSimpleName(dummyName));
		varDeclFrag.setInitializer(OperandType.fromDomType(param.getType()).dummyExpression());
		VariableDeclarationStatement dummyDecl = ast.newVariableDeclarationStatement(varDeclFrag);
		
		// make referencing expression		
		Expression paramExpr = ast.newSimpleName(param.getName().toString());
		Expression nestedExpr1 = exprBuilder1.build(paramExpr, ast.newSimpleName("unneeded"));
		Expression nestedExpr2 = exprBuilder2.build(nestedExpr1, ast.newSimpleName(dummyName));
		Statement stmt = stmtBuilder.build(nestedExpr2);
		Block methBody = methBodyBuilder.build(dummyDecl, stmt);
		
		// make method declaration
		MethodDeclaration methDecl = ast.newMethodDeclaration();
		methDecl.setName(ast.newSimpleName(methodName));
		methDecl.setReturnType2(ast.newPrimitiveType(PrimitiveType.VOID));
		methDecl.setBody(methBody);
		if (methBodyBuilder.requiresFinal()) {
			param.modifiers().add(ast.newModifier(Modifier.ModifierKeyword.FINAL_KEYWORD));
		}
		methDecl.parameters().add(param);
		
		return methDecl;
	}

}
