package edu.uiuc.astgen.methodrelationships;

import java.util.List;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.ReturnStatement;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;
import org.testorrery.FilteredCompositeGenerator;
import org.testorrery.IGenerator;
import org.testorrery.link.IIterationLink;
import org.testorrery.link.LinkCreator;

import edu.uiuc.astgen.ASTNodeGeneratorBase;
import edu.uiuc.astgen.nestedexpressions.NestedExpressionBuilder;
import edu.uiuc.astgen.nestedexpressions.NestedExpressionBuilderGenerator;
import edu.uiuc.astgen.nestedexpressions.NotNestedBuilder;
import edu.uiuc.astgen.nestedexpressions.OperandType;
import edu.uiuc.astgen.nestedexpressions.ParenthesesBuilder;
import edu.uiuc.astgen.statements.ExpressionStatementBuilder;
import edu.uiuc.astgen.statements.StatementWithExpressionBuilder;
import edu.uiuc.astgen.statements.StatementWithExpressionBuilderGenerator;

public class SingleClassMethodReferenceGenerator extends FilteredCompositeGenerator<TypeDeclaration> {

	private IGenerator<MethodDeclaration> methDeclGen;
	private MethodRelationshipBuilderGenerator methRelGen;
	private MethodReferenceBuilderGenerator methRefGen;
	private IGenerator<NestedExpressionBuilder> methExprGen;
	private IGenerator<NestedExpressionBuilder> nestedExprGen;
	private IGenerator<StatementWithExpressionBuilder> exprInStmtGen;
	
	private static AST ast = ASTNodeGeneratorBase.ast();
	
	public SingleClassMethodReferenceGenerator(IGenerator<MethodDeclaration> methDeclGen) {
		this.methDeclGen = methDeclGen;
	}
	
	@Override
	protected IIterationLink getInitializedIterator() {
		if (methRelGen == null) {
			methRelGen = new MethodRelationshipBuilderGenerator();
		}
		if (methRefGen == null) {
			methRefGen = new MethodReferenceBuilderGenerator();
		}
		if (methExprGen == null) {			
			methExprGen = NestedExpressionBuilderGenerator.parenNoParenAllOperandTypes();
		}
		if (nestedExprGen == null) {
			nestedExprGen = NestedExpressionBuilderGenerator.allOperandTypes();
		}
		if (exprInStmtGen == null) {
			exprInStmtGen = StatementWithExpressionBuilderGenerator.allExpressionTypes();
		}
		return LinkCreator.createCombinatoricLink(
				methDeclGen, 
				methRelGen, 
				methRefGen,
				methExprGen,
				nestedExprGen,
				exprInStmtGen);
	}
	
	@Override
	protected boolean iterationIsValid() {
		MethodDeclaration methDecl = (MethodDeclaration)ASTNode.copySubtree(ast, methDeclGen.current());
		MethodRelationshipBuilder methRelBuilder = methRelGen.current();
		MethodReferenceBuilder methRefBuilder = methRefGen.current();
		NestedExpressionBuilder methExprBuilder = methExprGen.current();
		NestedExpressionBuilder nestedExprBuilder = nestedExprGen.current();
		StatementWithExpressionBuilder exprInStmtBuilder = exprInStmtGen.current();		
		
		if (methDecl.parameters().size() == 0
			&& methRelBuilder.willRemoveParam()) {
			// disallow overloading with fewer parameters
			return false;
		}
		if (methRelBuilder.requiresSuper()) {
			// disallow overriding
			return false;
		}		
		if (methRefBuilder.isSuper()) {
			// disallow super.m() calls since this will only have one class
			return false;
		}
		if (!methExprBuilder.applicableTo(OperandType.fromDomType(methDecl.getReturnType2()))) {
			// disallow expressions that are not applicable to the method's return type
			return false;
		}
		if (nestedExprBuilder.isAssignment()) {
			// cannot assign to a method
			return false;
		}
		if (!nestedExprBuilder.applicableTo(methExprBuilder.outputType())) {
			// if output type of nested expression does not match the input type of the outer expression
			return false;
		}
		if (!exprInStmtBuilder.applicableTo(nestedExprBuilder.outputType())) {
			// if output type of outer expression does not match the input type of the statement
			return false;
		}
		if (methExprBuilder instanceof ParenthesesBuilder 
				&& exprInStmtBuilder instanceof ExpressionStatementBuilder) {
			// disallow parenthesized expression statement
			// (expr);
			return false;
		}
		if (methExprBuilder instanceof NotNestedBuilder 
				&& nestedExprBuilder instanceof ParenthesesBuilder
				&& exprInStmtBuilder instanceof ExpressionStatementBuilder) {
			// disallow parenthesized expression statement
			// (expr);
			return false;
		}
		return true;
	}

	@Override
	protected TypeDeclaration generateCurrent() {
		MethodDeclaration methDecl = (MethodDeclaration)ASTNode.copySubtree(ast, methDeclGen.current());
		MethodRelationshipBuilder methRelBuilder = methRelGen.current();
		MethodReferenceBuilder methRefBuilder = methRefGen.current();
		NestedExpressionBuilder methExprBuilder = methExprGen.current();
		NestedExpressionBuilder nestedExprBuilder = nestedExprGen.current();
		StatementWithExpressionBuilder exprInStmtBuilder = exprInStmtGen.current();
	
		TypeDeclaration toReturn = ast.newTypeDeclaration();
		toReturn.setName(ast.newSimpleName("A"));
		toReturn.bodyDeclarations().add(methDecl);
		toReturn.modifiers().add(ast.newModifier(Modifier.ModifierKeyword.PUBLIC_KEYWORD));
		
		String methodName = methDecl.getName().toString();
		String newMethodName = methodName + "Prime";
		String dummyVarName = "dummyVariable";		
		
		MethodDeclaration newMethDecl = methRelBuilder.build(newMethodName, methDecl, toReturn);
		Expression methRefExpr = methRefBuilder.build(methDecl, toReturn);
		Expression methExpr = methExprBuilder.build(methRefExpr, ast.newSimpleName("unneeded")); 
		Expression nestedExpr = nestedExprBuilder.build(methExpr, ast.newSimpleName(dummyVarName));
		Statement stmt = exprInStmtBuilder.build(nestedExpr);
		List newMethStmts = newMethDecl.getBody().statements();
		
		if (!methDecl.getReturnType2().toString().equals("void")) {
			VariableDeclarationFragment varDeclFrag = ast.newVariableDeclarationFragment();
			varDeclFrag.setName(ast.newSimpleName(dummyVarName));
			varDeclFrag.setInitializer(ast.newNullLiteral());
			VariableDeclarationStatement dummyVarDecl = ast.newVariableDeclarationStatement(varDeclFrag);
			dummyVarDecl.setType((Type)ASTNode.copySubtree(ast, methDecl.getReturnType2()));
			newMethStmts.add(dummyVarDecl);						
		}
		newMethStmts.add(stmt);
		
		if (!methDecl.getReturnType2().toString().equals("void")) {
			ReturnStatement methReturn = ast.newReturnStatement();
			methReturn.setExpression(OperandType.fromDomType(methDecl.getReturnType2()).dummyExpression());
			methDecl.getBody().statements().add(methReturn);
			
			ReturnStatement newMethReturn = ast.newReturnStatement();
			newMethReturn.setExpression(ast.newSimpleName(dummyVarName));
			newMethStmts.add(newMethReturn);
		}

		
		return toReturn;
	}

}
