package edu.uiuc.astgen.nestedexpressions;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ConditionalExpression;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.InfixExpression;
import org.eclipse.jdt.core.dom.Name;

import edu.uiuc.astgen.ASTNodeGeneratorBase;

/**
 * Enum of places in which a target expression can be located within a conditional expression
 */
public enum ConditionalOperandLocation {
	IF_EXPRESSION(new IfBuilder()),
	THEN_EXPRESSION(new ThenBuilder()),
	ELSE_EXPRESSION(new ElseBuilder()), 
	IF_AND_THEN(new ThenBuilder(), new IfBuilder()),
	IF_AND_ELSE(new ElseBuilder(), new IfBuilder()),
	ELSE_AND_THEN(new ElseBuilder(), new ThenBuilder()),
	ALL_THREE(new ElseBuilder(), new ThenBuilder(), new IfBuilder());
	
	private static final AST ast = ASTNodeGeneratorBase.ast();

	private static interface ConditionalBuilder {
		void build(
				ConditionalExpression condExpr, 
				Expression targetExpression, 
				Name dummyVariable, 
				OperandType operandType);			
	}
		
	private static class IfBuilder implements ConditionalBuilder {
		public void build(
				ConditionalExpression condExpr, 
				Expression targetExpression, 
				Name dummyVariable, 
				OperandType operandType) {
			
			if (operandType == OperandType.BOOLEAN) {
				condExpr.setExpression((Expression)ASTNode.copySubtree(ast, targetExpression));
			}
			else {
				// build == expression so that the "if" part will always compile
				InfixExpression eqExp = ast.newInfixExpression();
				eqExp.setOperator(InfixExpression.Operator.EQUALS);
				eqExp.setLeftOperand((Expression)ASTNode.copySubtree(ast, targetExpression));
				eqExp.setRightOperand((Expression)ASTNode.copySubtree(ast, dummyVariable)); 
				condExpr.setExpression(eqExp);
			}			
		}
	}
	
	private static class ThenBuilder implements ConditionalBuilder {
		public void build(ConditionalExpression condExpr, Expression targetExpression, Name dummyVariable, OperandType operandType) {
			condExpr.setThenExpression((Expression)ASTNode.copySubtree(ast, targetExpression));
			if (operandType != OperandType.BOOLEAN) {
				InfixExpression eqExp = ast.newInfixExpression();
				eqExp.setOperator(InfixExpression.Operator.EQUALS);
				eqExp.setLeftOperand((Expression)ASTNode.copySubtree(ast, dummyVariable));
				eqExp.setRightOperand((Expression)ASTNode.copySubtree(ast, dummyVariable)); 
				condExpr.setExpression(eqExp);
			}
		}
	}
	
	private static class ElseBuilder implements ConditionalBuilder {
		public void build(ConditionalExpression condExpr, Expression targetExpression, Name dummyVariable, OperandType operandType) {
			condExpr.setElseExpression((Expression)ASTNode.copySubtree(ast, targetExpression));
			if (operandType != OperandType.BOOLEAN) {
				InfixExpression eqExp = ast.newInfixExpression();
				eqExp.setOperator(InfixExpression.Operator.NOT_EQUALS);
				eqExp.setLeftOperand((Expression)ASTNode.copySubtree(ast, dummyVariable));
				eqExp.setRightOperand((Expression)ASTNode.copySubtree(ast, dummyVariable)); 
				condExpr.setExpression(eqExp);
			}
		}
	}
	
	private ConditionalBuilder[] builders;
	private ConditionalOperandLocation(ConditionalBuilder... builders) {
		this.builders = builders;
	}
	
	public ConditionalExpression build(Expression targetExpression, Name dummyVariable, OperandType operandType) {
		ConditionalExpression condExpr = ast.newConditionalExpression();
		condExpr.setExpression((Expression)ASTNode.copySubtree(ast, dummyVariable));
		condExpr.setThenExpression((Expression)ASTNode.copySubtree(ast, dummyVariable));
		condExpr.setElseExpression((Expression)ASTNode.copySubtree(ast, dummyVariable));
		
		for (ConditionalBuilder builder : builders) {
			builder.build(condExpr, targetExpression, dummyVariable, operandType);
		}
		return condExpr;
	}
	
}
