package edu.uiuc.astgen.nestedexpressions;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Assignment;
import org.eclipse.jdt.core.dom.CastExpression;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.InfixExpression;
import org.eclipse.jdt.core.dom.PostfixExpression;
import org.eclipse.jdt.core.dom.PrefixExpression;
import org.eclipse.jdt.core.dom.Type;

import edu.uiuc.astgen.ASTNodeGeneratorBase;

/**
 * Enum of Expression operand types and their associated operators.
 */
public enum OperandType {
	VOID(null,
			new Assignment.Operator[] {},
			new InfixExpression.Operator[] {},
			new PrefixExpression.Operator[] {},
			new PostfixExpression.Operator[] {}),
	OBJECT(getObjectExpression(),
			new Assignment.Operator[] {
					Assignment.Operator.ASSIGN},
			new InfixExpression.Operator[] {
					InfixExpression.Operator.EQUALS, 
					InfixExpression.Operator.NOT_EQUALS},
			new PrefixExpression.Operator[] {},
			new PostfixExpression.Operator[] {}),					
	NUMERIC(ASTNodeGeneratorBase.ast().newNumberLiteral("0"),
			new Assignment.Operator[] {
					Assignment.Operator.ASSIGN, 
					Assignment.Operator.BIT_AND_ASSIGN,
					Assignment.Operator.BIT_OR_ASSIGN,
					Assignment.Operator.BIT_XOR_ASSIGN,
					Assignment.Operator.DIVIDE_ASSIGN,
					Assignment.Operator.LEFT_SHIFT_ASSIGN,
					Assignment.Operator.MINUS_ASSIGN,
					Assignment.Operator.PLUS_ASSIGN,
					Assignment.Operator.REMAINDER_ASSIGN,
					Assignment.Operator.RIGHT_SHIFT_SIGNED_ASSIGN,
					Assignment.Operator.RIGHT_SHIFT_UNSIGNED_ASSIGN,
					Assignment.Operator.TIMES_ASSIGN},
			new InfixExpression.Operator[] {
					InfixExpression.Operator.EQUALS, 
					InfixExpression.Operator.NOT_EQUALS,
					InfixExpression.Operator.AND,
					InfixExpression.Operator.DIVIDE,
					InfixExpression.Operator.GREATER,
					InfixExpression.Operator.GREATER_EQUALS,
					InfixExpression.Operator.LEFT_SHIFT,
					InfixExpression.Operator.LESS,
					InfixExpression.Operator.LESS_EQUALS,
					InfixExpression.Operator.MINUS,
					InfixExpression.Operator.OR,
					InfixExpression.Operator.PLUS,
					InfixExpression.Operator.REMAINDER,
					InfixExpression.Operator.RIGHT_SHIFT_SIGNED,
					InfixExpression.Operator.RIGHT_SHIFT_UNSIGNED,
					InfixExpression.Operator.TIMES,
					InfixExpression.Operator.XOR},
			new PrefixExpression.Operator[] {
					PrefixExpression.Operator.COMPLEMENT,
					PrefixExpression.Operator.DECREMENT,
					PrefixExpression.Operator.INCREMENT,
					PrefixExpression.Operator.MINUS,
					PrefixExpression.Operator.PLUS},
			new PostfixExpression.Operator[] {
					PostfixExpression.Operator.DECREMENT,
					PostfixExpression.Operator.INCREMENT}),
	BOOLEAN(ASTNodeGeneratorBase.ast().newBooleanLiteral(false),
			new Assignment.Operator[] {
					Assignment.Operator.ASSIGN,
					Assignment.Operator.BIT_AND_ASSIGN,
					Assignment.Operator.BIT_OR_ASSIGN,
					Assignment.Operator.BIT_XOR_ASSIGN},
			new InfixExpression.Operator[] {
					InfixExpression.Operator.EQUALS, 
					InfixExpression.Operator.NOT_EQUALS,
					InfixExpression.Operator.AND,
					InfixExpression.Operator.CONDITIONAL_AND,
					InfixExpression.Operator.CONDITIONAL_OR,
					InfixExpression.Operator.OR,
					InfixExpression.Operator.XOR},
			new PrefixExpression.Operator[] {
					PrefixExpression.Operator.NOT,
					PrefixExpression.Operator.COMPLEMENT},
			new PostfixExpression.Operator[] {});
	
	private Expression dummyExpression;
	private Assignment.Operator[] assignmentOps;
	private InfixExpression.Operator[] infixOps;
	private PrefixExpression.Operator[] prefixOps;
	private PostfixExpression.Operator[] postfixOps;
	
	public Expression dummyExpression() {
		return (Expression)ASTNode.copySubtree(ASTNodeGeneratorBase.ast(), dummyExpression);
	}
	public Assignment.Operator[] getAssignmentOperators() {
		return assignmentOps;
	}
	public InfixExpression.Operator[] getInfixOperators() {
		return infixOps;
	}
	public PrefixExpression.Operator[] getPrefixOperators() {
		return prefixOps;
	}
	public PostfixExpression.Operator[] getPostfixOperators() {
		return postfixOps;
	}
	
	private OperandType(
			Expression dummyExpression,
			Assignment.Operator[] assignmentOps,
			InfixExpression.Operator[] infixOps,
			PrefixExpression.Operator[] prefixOps,
			PostfixExpression.Operator[] postfixOps) {
		this.dummyExpression = dummyExpression;
		this.assignmentOps = assignmentOps;
		this.infixOps = infixOps;
		this.postfixOps = postfixOps;
		this.prefixOps = prefixOps;
	}
	
	
	public static OperandType fromDomType(Type domType) {
		String typeName = domType.toString();
		if (
				typeName.equals("Byte") ||
				typeName.equals("byte") ||
				typeName.equals("Integer") || 
				typeName.equals("int") ||
				typeName.equals("Long") ||
				typeName.equals("long")
				// TODO: more...
				) {
			return NUMERIC;
		}
		
		if (
				typeName.equals("Boolean") ||
				typeName.equals("boolean")) {
			return BOOLEAN;
		} 
		
		if (typeName.equals("void")) {
			return VOID;
		}
		
		if (typeName.equals("Object")) {
			return OBJECT;
		}
		
		throw new RuntimeException("Unknown DOM type: " + typeName);
	}
	
	private static Expression getObjectExpression() {
		AST ast = ASTNodeGeneratorBase.ast();
		// (Object)null
		CastExpression cast = ast.newCastExpression();
		cast.setType(ast.newSimpleType(ast.newSimpleName("Object")));
		cast.setExpression(ast.newNullLiteral());
		return cast;
	}
	
}
