package edu.uiuc.astgen.nestedexpressions;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.Type;
import org.testorrery.Chain;
import org.testorrery.CompositeGenerator;
import org.testorrery.IGenerator;
import org.testorrery.link.IIterationLink;
import org.testorrery.link.LinkCreator;

import edu.uiuc.astgen.ASTNodeGeneratorBase;
import edu.uiuc.astgen.methodrelationships.SingleClassMethodReferenceGenerator;

/**
 * Generates type-correct {@link TypedFieldExpressionSpec} when given a field declaration. 
 * 
 * This TypedFieldExpressionSpec stuff is a nasty hack to get expressions to be type-correct.
 * See {@link SingleClassMethodReferenceGenerator#iterationIsValid for a better way to do it.
 */
public class TypedFieldExpressionSpecGenerator extends
		CompositeGenerator<TypedFieldExpressionSpec> {
	
	private static Map<String, Expression> initializingExpresionMap;
	private static Expression defaultInitializingExpression;
	private static Map<String, NestedExpressionBuilderGenerator> nestedExpressionBuilderGeneratorMap;
	private static NestedExpressionBuilderGenerator defaultNestedExpressionBuilderGen;
	private static AST ast = ASTNodeGeneratorBase.ast();
	
	static {
		defaultInitializingExpression = ast.newNullLiteral();		
		initializingExpresionMap = new HashMap<String, Expression>();	
		setExprForTypes(ast.newNumberLiteral("0"), 
				"Integer", 
				"int",
				"Byte",
				"byte");		
		setExprForTypes(ast.newNumberLiteral("0L"), 
				"Long",
				"long");
		setExprForTypes(ast.newBooleanLiteral(false), 
				"Boolean", 
				"boolean");
		// TODO: more types
		
		defaultNestedExpressionBuilderGen = new NestedExpressionBuilderGenerator(OperandType.OBJECT);
		nestedExpressionBuilderGeneratorMap = new HashMap<String, NestedExpressionBuilderGenerator>();
		setBuilderGenForTypes(new NestedExpressionBuilderGenerator(OperandType.NUMERIC),
				"Byte",
				"byte",
				"Integer", 
				"int",		
				"Long",
				"long");
		setBuilderGenForTypes(new NestedExpressionBuilderGenerator(OperandType.BOOLEAN),
				"Boolean", 
				"boolean");		
	}
	
	private static void setExprForTypes(Expression expression, String... typeNames) {
		for (String typeName : typeNames) {
			initializingExpresionMap.put(typeName, expression);
		}		
	}
	
	private static void setBuilderGenForTypes(NestedExpressionBuilderGenerator builderGen, String... typeNames) {
		for (String typeName : typeNames) {
			nestedExpressionBuilderGeneratorMap.put(typeName, builderGen);
		}
	}
	
	private FieldDeclaration fieldDecl;
	private Expression initializingExpr;
	private IGenerator<NestedExpressionBuilder> nestedExprBuilderGen;

	public TypedFieldExpressionSpecGenerator(FieldDeclaration fieldDecl) {
		this.fieldDecl = fieldDecl;
		this.initializingExpr = createInitializingExpression(fieldDecl.getType());
		this.nestedExprBuilderGen = createNestedExprBuilderGen(fieldDecl.getType());
	}


	public static IGenerator<TypedFieldExpressionSpec> newGenerator(IGenerator<FieldDeclaration> fieldDeclGen) {
		Chain<TypedFieldExpressionSpec> seqGen = new Chain<TypedFieldExpressionSpec>();
		for (FieldDeclaration fieldDecl : fieldDeclGen) {
			seqGen.add(new TypedFieldExpressionSpecGenerator(fieldDecl));
		}
		return seqGen;
	}

	@Override
	protected IIterationLink getInitializedIterator() {
		return LinkCreator.createCombinatoricLink(nestedExprBuilderGen);
	}

	@Override
	protected TypedFieldExpressionSpec generateCurrent() {
		NestedExpressionBuilder nestedExprBuilder = nestedExprBuilderGen.current();
		return new TypedFieldExpressionSpec(fieldDecl, nestedExprBuilder, initializingExpr);
	}
	
	private Expression createInitializingExpression(Type type) {
		Expression initExpr = (Expression)ASTNode.copySubtree(ast, initializingExpresionMap.get(type.toString()));
		if (initExpr == null) {
			return defaultInitializingExpression;
		}
		return initExpr;
	}
	
	private IGenerator<NestedExpressionBuilder> createNestedExprBuilderGen(Type type) {
		NestedExpressionBuilderGenerator builderGen = nestedExpressionBuilderGeneratorMap.get(type.toString());
		if (builderGen == null) {
			return defaultNestedExpressionBuilderGen;
		}
		return builderGen;
	}

}
