package edu.uiuc.astgen.test.classrelationships;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.junit.*;

import edu.uiuc.astgen.basic.ClassGenerator;
import edu.uiuc.astgen.classrelationships.RelatedPair;
import edu.uiuc.astgen.classrelationships.ReferenceType;
import edu.uiuc.astgen.test.ASTNodeGeneratorTestUtil;

public class ReferenceTypeTest {
	TypeDeclaration a, b;
	
	@Before
	public void setUp() {
		a = new ClassGenerator("A").next();
		b = new ClassGenerator("B").next();
	}

	@Test
	public void testNoReference() {
		runTest(
				ReferenceType.NO_REFERENCE,
				"class A {\n}\n",
				"class B {\n}\n");
	}
	
	@Test
	public void testFieldType() {
		runTest(
				ReferenceType.FIELD_TYPE,
				"class A {\n}\n",
				"class B {\n  public A dummyField_B;\n}\n");
	}
	@Test
	public void testLocalVariableType() {
		runTest(
				ReferenceType.LOCAL_VARIABLE_TYPE,
				"class A {\n}\n",
				"class B {\n  public void dummyMethod_B(){\n    A theA;\n  }\n}\n");
	}
	@Test
	public void testParameterType() {
		runTest(
				ReferenceType.PARAMETER_TYPE,
				"class A {\n}\n",
				"class B {\n  public void dummyMethod_B(  A arg0){\n  }\n}\n");
	}
	@Test
	public void testConstructorParameterType() {
		runTest(
				ReferenceType.CONSTRUCTOR_PARAMETER_TYPE,
				"class A {\n}\n",
				"class B {\n  public B(  A arg0){\n  }\n}\n");
	}
	@Test
	public void testReturnType() {
		runTest(
				ReferenceType.RETURN_TYPE,
				"class A {\n}\n",
				"class B {\n  public A dummyMethod_B(){\n    return null;\n  }\n}\n");
	}
	@Test
	public void testStaticMethodCall() {
		runTest(
				ReferenceType.STATIC_METHOD_CALL,
				"class A {\n  public static void dummyMethod_A(){\n  }\n}\n",
				"class B {\n  public void dummyMethod_B(){\n    A.dummyMethod_A();\n  }\n}\n");
	}
	@Test
	public void testCastExpression() {
		runTest(
				ReferenceType.CAST_EXPRESSION,
				"class A {\n}\n",
				"class B {\n  public void dummyMethod_B(){\n    Object dummyVariable_dummyMethod_B=(A)null;\n  }\n}\n");
	}
	@Test
	public void testInstanceofExpression() {
		runTest(
				ReferenceType.INSTANCEOF_EXPRESSION,
				"class A {\n}\n",
				"class B {\n  public void dummyMethod_B(){\n    Object dummyVariable_dummyMethod_B=null instanceof A;\n  }\n}\n");
	}	
	@Test
	public void testForLoopVariableType() {
		runTest(
				ReferenceType.FOR_LOOP_VARIABLE_TYPE,
				"class A {\n}\n",
				"class B {\n  public void dummyMethod_B(){\n    for (A i; ; ) {\n    }\n  }\n}\n");
	}
	@Test
	public void testCatchClauseParameterType() {
		runTest(
				ReferenceType.CATCH_CLAUSE_PARAMETER_TYPE,
				"class A {\n}\n",
				"class B {\n  public void dummyMethod_B(){\n    try {\n    }\n catch (    A e) {\n    }\n  }\n}\n");
	}
	@Test
	public void testThrowsClauseType() {
		runTest(
				ReferenceType.THROWS_CLAUSE_TYPE,
				"class A {\n}\n",
				"class B {\n  public void dummyMethod_B() throws A {\n  }\n}\n");
	}
	
	@Test
	public void testConstructorCallExpression() {
		AST ast = ASTNodeGeneratorTestUtil.getSampleAST();
		
		// add a constructor and parameters to test constructor call generation
		SingleVariableDeclaration param1 = ast.newSingleVariableDeclaration();
		param1.setName(ast.newSimpleName("arg0"));
		param1.setType(ast.newSimpleType(ast.newSimpleName("Object")));
		SingleVariableDeclaration param2 = ast.newSingleVariableDeclaration();
		param2.setName(ast.newSimpleName("arg1"));
		param2.setType(ast.newSimpleType(ast.newSimpleName("Object")));		
		MethodDeclaration methDecl = ast.newMethodDeclaration();
		methDecl.setName(ast.newSimpleName(a.getName().toString()));	
		methDecl.setConstructor(true);
		methDecl.parameters().add(param1);
		methDecl.parameters().add(param2);
		methDecl.setBody(ast.newBlock());
		
		a.bodyDeclarations().add(methDecl);
		
		runTest(
				ReferenceType.CONSTRUCTOR_CALL_EXPRESSION, 
				"class A {\n  A(  Object arg0,  Object arg1){\n  }\n}\n",
				"class B {\n  public void dummyMethod_B(){\n    Object dummyVariable_dummyMethod_B=new A(null,null);\n  }\n}\n");
	}
	
	@Test
	public void testTypeLiteralExpression() {		
		runTest(
				ReferenceType.TYPE_LITERAL_EXPRESSION, 
				"class A {\n}\n",
				"class B {\n  public void dummyMethod_B(){\n    Object dummyVariable_dummyMethod_B=A.class;\n  }\n}\n");
	}
	
	@Test
	public void testStaticFieldReferenceExpression() {
		runTest(
				ReferenceType.STATIC_FIELD_REFERENCE_EXPRESSION, 
				"class A {\n  public static Object dummyField_A;\n}\n",
				"class B {\n  public void dummyMethod_B(){\n    Object dummyVariable_dummyMethod_B=A.dummyField_A;\n  }\n}\n");
	}
	
	@Test
	public void testStaticMethodCallExpression() {
		runTest(
				ReferenceType.STATIC_METHOD_CALL, 
				"class A {\n  public static void dummyMethod_A(){\n  }\n}\n",
				"class B {\n  public void dummyMethod_B(){\n    A.dummyMethod_A();\n  }\n}\n");
	}
	
	@Test
	public void testArrayCreationExpression() {
		runTest(
				ReferenceType.ARRAY_CREATION_EXPRESSION, 
				"class A {\n}\n",
				"class B {\n  public void dummyMethod_B(){\n    Object dummyVariable_dummyMethod_B=new A[]{};\n  }\n}\n");
	}
	
	@Test
	public void testQualifiedThisExpression() {
		runTest(
				ReferenceType.QUALIFIED_THIS_EXPRESSION, 
				"class A {\n}\n",
				"class B {\n  public void dummyMethod_B(){\n    Object dummyVariable_dummyMethod_B=A.this;\n  }\n}\n");
	}
		
	void runTest(ReferenceType type, String expectedA, String expectedB) {		
		RelationshipSpecificationTest.testSpec(type, a, b, expectedA, expectedB);			
	}
}
